# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import os
import mock
import json
import logging
import argparse
import tempfile
import unittest2
from collections import namedtuple

from tests import base
from tests.base import BaseCLITestCase

from st2client.shell import Shell
from st2client import models
from st2client.utils import httpclient
from st2client.commands import resource
from st2client.commands.resource import ResourceViewCommand

__all__ = ["TestResourceCommand", "ResourceViewCommandTestCase"]


LOG = logging.getLogger(__name__)


class TestCommands(base.BaseCLITestCase):
    def __init__(self, *args, **kwargs):
        super(TestCommands, self).__init__(*args, **kwargs)
        self.shell = Shell()

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps({}), 404, "NOT FOUND")
        ),
    )
    def test_all_resources_get_multi(self):
        # 1. Verify that st2 <resource> get <id 1> ... <id n> notation works for all get commands
        resources = [
            ("action", models.Action),
            ("action-alias", models.ActionAlias),
            ("rule", models.Rule),
            ("sensor", models.Sensor),
            ("pack", models.Pack),
            ("execution", models.Execution),
            ("key", models.KeyValuePair),
            ("webhook", models.Webhook),
            ("trigger-instance", models.TriggerInstance),
            ("trigger", models.TriggerType),
            ("apikey", models.ApiKey),
            ("inquiry", models.Inquiry),
            ("policy", models.Policy),
            ("policy-type", models.PolicyType),
            ("timer", models.Timer),
            ("trace", models.Trace),
            ("runner", models.RunnerType),
            ("rule-enforcement", models.RuleEnforcement),
            ("role", models.Role),
            ("role-assignment", models.UserRoleAssignment),
        ]

        for command_name, resource_ in resources:
            display_name = resource_.get_display_name()

            self._reset_output_streams()
            return_code = self.shell.run([command_name, "get", "id1", "id2", "id3"])
            self.assertEqual(return_code, 0)

            stdout = self.stdout.getvalue()
            self.assertTrue('%s "id1" is not found.' % (display_name) in stdout)
            self.assertTrue('%s "id2" is not found.' % (display_name) in stdout)
            self.assertTrue('%s "id3" is not found.' % (display_name) in stdout)
            self._reset_output_streams()

        # 2. When a single id is provided, command should return non-zero in case resource is not
        # found
        for command_name, resource_ in resources:
            display_name = resource_.get_display_name()

            self._reset_output_streams()
            return_code = self.shell.run([command_name, "get", "id3"])
            self.assertEqual(return_code, 1)

            stdout = self.stdout.getvalue()
            self.assertTrue('%s "id3" is not found.' % (display_name) in stdout)
            self._reset_output_streams()


class TestResourceCommand(unittest2.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestResourceCommand, self).__init__(*args, **kwargs)
        self.parser = argparse.ArgumentParser()
        self.subparsers = self.parser.add_subparsers()
        self.branch = resource.ResourceBranch(
            base.FakeResource, "Test Command", base.FakeApp(), self.subparsers
        )

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(base.RESOURCES), 200, "OK")
        ),
    )
    def test_command_list(self):
        args = self.parser.parse_args(["fakeresource", "list"])
        self.assertEqual(args.func, self.branch.commands["list"].run_and_print)
        instances = self.branch.commands["list"].run(args)
        actual = [instance.serialize() for instance in instances]
        expected = json.loads(json.dumps(base.RESOURCES))
        self.assertListEqual(actual, expected)

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse("", 500, "INTERNAL SERVER ERROR")
        ),
    )
    def test_command_list_failed(self):
        args = self.parser.parse_args(["fakeresource", "list"])
        self.assertRaises(Exception, self.branch.commands["list"].run, args)

    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(return_value=None)
    )
    @mock.patch.object(
        models.ResourceManager,
        "get_by_id",
        mock.MagicMock(return_value=base.FakeResource(**base.RESOURCES[0])),
    )
    def test_command_get_by_id(self):
        args = self.parser.parse_args(["fakeresource", "get", "123"])
        self.assertEqual(args.func, self.branch.commands["get"].run_and_print)
        instances = self.branch.commands["get"].run(args)
        actual = instances[0].serialize()
        expected = json.loads(json.dumps(base.RESOURCES[0]))
        self.assertEqual(actual, expected)

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(base.RESOURCES[0]), 200, "OK")
        ),
    )
    def test_command_get_single(self):
        args = self.parser.parse_args(["fakeresource", "get", "abc"])
        self.assertEqual(args.func, self.branch.commands["get"].run_and_print)
        instances = self.branch.commands["get"].run(args)
        self.assertEqual(len(instances), 1)
        instance = instances[0]
        actual = instance.serialize()
        expected = json.loads(json.dumps(base.RESOURCES[0]))
        self.assertEqual(actual, expected)

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            side_effect=[
                base.FakeResponse(json.dumps(base.RESOURCES[0]), 200, "OK"),
                base.FakeResponse(json.dumps(base.RESOURCES[1]), 200, "OK"),
            ]
        ),
    )
    def test_command_get_multiple(self):
        args = self.parser.parse_args(["fakeresource", "get", "abc", "def"])
        self.assertEqual(args.func, self.branch.commands["get"].run_and_print)
        instances = self.branch.commands["get"].run(args)
        self.assertEqual(len(instances), 2)

        actual = instances[0].serialize()
        expected = json.loads(json.dumps(base.RESOURCES[0]))
        self.assertEqual(actual, expected)

        actual = instances[1].serialize()
        expected = json.loads(json.dumps(base.RESOURCES[1]))
        self.assertEqual(actual, expected)

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(return_value=base.FakeResponse("", 404, "NOT FOUND")),
    )
    def test_command_get_404(self):
        args = self.parser.parse_args(["fakeresource", "get", "cba"])
        self.assertEqual(args.func, self.branch.commands["get"].run_and_print)
        self.assertRaises(
            resource.ResourceNotFoundError, self.branch.commands["get"].run, args
        )

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse("", 500, "INTERNAL SERVER ERROR")
        ),
    )
    def test_command_get_failed(self):
        args = self.parser.parse_args(["fakeresource", "get", "cba"])
        self.assertRaises(Exception, self.branch.commands["get"].run, args)

    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(base.RESOURCES[0]), 200, "OK")
        ),
    )
    def test_command_create(self):
        instance = base.FakeResource(name="abc")
        fd, path = tempfile.mkstemp(suffix=".json")
        try:
            with open(path, "a") as f:
                f.write(json.dumps(instance.serialize(), indent=4))
            args = self.parser.parse_args(["fakeresource", "create", path])
            self.assertEqual(args.func, self.branch.commands["create"].run_and_print)
            instance = self.branch.commands["create"].run(args)
            actual = instance.serialize()
            expected = json.loads(json.dumps(base.RESOURCES[0]))
            self.assertEqual(actual, expected)
        finally:
            os.close(fd)
            os.unlink(path)

    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse("", 500, "INTERNAL SERVER ERROR")
        ),
    )
    def test_command_create_failed(self):
        instance = base.FakeResource(name="abc")
        fd, path = tempfile.mkstemp(suffix=".json")
        try:
            with open(path, "a") as f:
                f.write(json.dumps(instance.serialize(), indent=4))
            args = self.parser.parse_args(["fakeresource", "create", path])
            self.assertRaises(Exception, self.branch.commands["create"].run, args)
        finally:
            os.close(fd)
            os.unlink(path)

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(
                json.dumps([base.RESOURCES[0]]), 200, "OK", {}
            )
        ),
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "put",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(base.RESOURCES[0]), 200, "OK")
        ),
    )
    def test_command_update(self):
        instance = base.FakeResource(id="123", name="abc")
        fd, path = tempfile.mkstemp(suffix=".json")
        try:
            with open(path, "a") as f:
                f.write(json.dumps(instance.serialize(), indent=4))
            args = self.parser.parse_args(["fakeresource", "update", "123", path])
            self.assertEqual(args.func, self.branch.commands["update"].run_and_print)
            instance = self.branch.commands["update"].run(args)
            actual = instance.serialize()
            expected = json.loads(json.dumps(base.RESOURCES[0]))
            self.assertEqual(actual, expected)
        finally:
            os.close(fd)
            os.unlink(path)

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps([base.RESOURCES[0]]), 200, "OK")
        ),
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "put",
        mock.MagicMock(
            return_value=base.FakeResponse("", 500, "INTERNAL SERVER ERROR")
        ),
    )
    def test_command_update_failed(self):
        instance = base.FakeResource(id="123", name="abc")
        fd, path = tempfile.mkstemp(suffix=".json")
        try:
            with open(path, "a") as f:
                f.write(json.dumps(instance.serialize(), indent=4))
            args = self.parser.parse_args(["fakeresource", "update", "123", path])
            self.assertRaises(Exception, self.branch.commands["update"].run, args)
        finally:
            os.close(fd)
            os.unlink(path)

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps([base.RESOURCES[0]]), 200, "OK")
        ),
    )
    def test_command_update_id_mismatch(self):
        instance = base.FakeResource(id="789", name="abc")
        fd, path = tempfile.mkstemp(suffix=".json")
        try:
            with open(path, "a") as f:
                f.write(json.dumps(instance.serialize(), indent=4))
            args = self.parser.parse_args(["fakeresource", "update", "123", path])
            self.assertRaises(Exception, self.branch.commands["update"].run, args)
        finally:
            os.close(fd)
            os.unlink(path)

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(
                json.dumps([base.RESOURCES[0]]), 200, "OK", {}
            )
        ),
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "delete",
        mock.MagicMock(return_value=base.FakeResponse("", 204, "NO CONTENT")),
    )
    def test_command_delete(self):
        args = self.parser.parse_args(["fakeresource", "delete", "abc"])
        self.assertEqual(args.func, self.branch.commands["delete"].run_and_print)
        self.branch.commands["delete"].run(args)

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(return_value=base.FakeResponse("", 404, "NOT FOUND")),
    )
    def test_command_delete_404(self):
        args = self.parser.parse_args(["fakeresource", "delete", "cba"])
        self.assertEqual(args.func, self.branch.commands["delete"].run_and_print)
        self.assertRaises(
            resource.ResourceNotFoundError, self.branch.commands["delete"].run, args
        )

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps([base.RESOURCES[0]]), 200, "OK")
        ),
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "delete",
        mock.MagicMock(
            return_value=base.FakeResponse("", 500, "INTERNAL SERVER ERROR")
        ),
    )
    def test_command_delete_failed(self):
        args = self.parser.parse_args(["fakeresource", "delete", "cba"])
        self.assertRaises(Exception, self.branch.commands["delete"].run, args)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_id",
        mock.MagicMock(return_value=base.FakeResource(**base.RESOURCES[0])),
    )
    def test_command_get_unicode_primary_key(self):
        args = self.parser.parse_args(
            ["fakeresource", "get", "examples.test_rule_utf8_náme"]
        )
        self.assertEqual(args.func, self.branch.commands["get"].run_and_print)
        instances = self.branch.commands["get"].run(args)
        actual = instances[0].serialize()
        expected = json.loads(json.dumps(base.RESOURCES[0]))
        self.assertEqual(actual, expected)


class ResourceViewCommandTestCase(unittest2.TestCase):
    def setUp(self):
        ResourceViewCommand.display_attributes = []

    def test_get_include_attributes(self):
        cls = namedtuple("Args", "attr")

        args = cls(attr=[])
        result = ResourceViewCommand._get_include_attributes(args=args)
        self.assertEqual(result, [])

        args = cls(attr=["result"])
        result = ResourceViewCommand._get_include_attributes(args=args)
        self.assertEqual(result, ["result"])

        args = cls(attr=["result", "trigger_instance"])
        result = ResourceViewCommand._get_include_attributes(args=args)
        self.assertEqual(result, ["result", "trigger_instance"])

        args = cls(attr=["result.stdout"])
        result = ResourceViewCommand._get_include_attributes(args=args)
        self.assertEqual(result, ["result.stdout"])

        args = cls(attr=["result.stdout", "result.stderr"])
        result = ResourceViewCommand._get_include_attributes(args=args)
        self.assertEqual(result, ["result.stdout", "result.stderr"])

        args = cls(attr=["result.stdout", "trigger_instance.id"])
        result = ResourceViewCommand._get_include_attributes(args=args)
        self.assertEqual(result, ["result.stdout", "trigger_instance.id"])

        ResourceViewCommand.display_attributes = ["id", "status"]

        args = cls(attr=[])
        result = ResourceViewCommand._get_include_attributes(args=args)
        self.assertEqual(set(result), set(["id", "status"]))

        args = cls(attr=["trigger_instance"])
        result = ResourceViewCommand._get_include_attributes(args=args)
        self.assertEqual(set(result), set(["trigger_instance"]))

        args = cls(attr=["all"])
        result = ResourceViewCommand._get_include_attributes(args=args)
        self.assertEqual(result, None)


class CommandsHelpStringTestCase(BaseCLITestCase):
    """
    Test case which verifies that all the commands support -h / --help flag.
    """

    capture_output = True

    # TODO: Automatically iterate all the available commands
    COMMANDS = [
        # action
        ["action", "list"],
        ["action", "get"],
        ["action", "create"],
        ["action", "update"],
        ["action", "delete"],
        ["action", "enable"],
        ["action", "disable"],
        ["action", "execute"],
        # execution
        ["execution", "cancel"],
        ["execution", "pause"],
        ["execution", "resume"],
        ["execution", "tail"],
    ]

    def test_help_command_line_arg_works_for_supported_commands(self):
        shell = Shell()

        for command in self.COMMANDS:
            # First test longhang notation
            argv = command + ["--help"]

            try:
                result = shell.run(argv)
            except SystemExit as e:
                self.assertEqual(e.code, 0)
            else:
                self.assertEqual(result, 0)

            stdout = self.stdout.getvalue()

            self.assertIn("usage:", stdout)
            self.assertIn(" ".join(command), stdout)
            # self.assertIn('positional arguments:', stdout)
            self.assertIn("optional arguments:", stdout)

            # Reset stdout and stderr after each iteration
            self._reset_output_streams()

            # Then shorthand notation
            argv = command + ["-h"]

            try:
                result = shell.run(argv)
            except SystemExit as e:
                self.assertEqual(e.code, 0)
            else:
                self.assertEqual(result, 0)

            stdout = self.stdout.getvalue()

            self.assertIn("usage:", stdout)
            self.assertIn(" ".join(command), stdout)
            # self.assertIn('positional arguments:', stdout)
            self.assertIn("optional arguments:", stdout)

            # Verify that the actual help usage string was triggered and not the invalid
            # "too few arguments" which would indicate command doesn't actually correctly handle
            # --help flag
            self.assertNotIn("too few arguments", stdout)

            self._reset_output_streams()



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import json
import mock
import logging

from tests import base

from st2client import shell
from st2client import models
from st2client.utils import httpclient

LOG = logging.getLogger(__name__)

RUNNER1 = {
    "enabled": True,
    "runner_parameters": {
        "bool": {"type": "boolean"},
        "int": {"type": "integer"},
        "float": {"type": "number"},
        "json": {"type": "object"},
        "list": {"type": "array"},
        "str": {"type": "string"},
    },
    "name": "mock-runner1",
}

ACTION1 = {
    "ref": "mockety.mock1",
    "runner_type": "mock-runner1",
    "name": "mock1",
    "parameters": {},
    "enabled": True,
    "entry_point": "",
    "pack": "mockety",
}

RUNNER2 = {
    "enabled": True,
    "runner_parameters": {},
    "name": "mock-runner2",
}

ACTION2 = {
    "ref": "mockety.mock2",
    "runner_type": "mock-runner2",
    "name": "mock2",
    "parameters": {
        "bool": {"type": "boolean"},
        "int": {"type": "integer"},
        "float": {"type": "number"},
        "json": {"type": "object"},
        "list": {"type": "array"},
        "str": {"type": "string"},
    },
    "enabled": True,
    "entry_point": "",
    "pack": "mockety",
}

LIVE_ACTION = {
    "action": "mockety.mock",
    "status": "complete",
    "result": {"stdout": "non-empty"},
}


def get_by_name(name, **kwargs):
    if name == "mock-runner1":
        return models.RunnerType(**RUNNER1)
    if name == "mock-runner2":
        return models.RunnerType(**RUNNER2)


def get_by_ref(**kwargs):
    ref = kwargs.get("ref_or_id", None)

    if not ref:
        raise Exception('Actions must be referred to by "ref".')

    if ref == "mockety.mock1":
        return models.Action(**ACTION1)
    if ref == "mockety.mock2":
        return models.Action(**ACTION2)


class ActionCommandTestCase(base.BaseCLITestCase):
    def __init__(self, *args, **kwargs):
        super(ActionCommandTestCase, self).__init__(*args, **kwargs)
        self.shell = shell.Shell()

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_runner_param_bool_conversion(self):
        self.shell.run(["run", "mockety.mock1", "bool=false"])
        expected = {
            "action": "mockety.mock1",
            "user": None,
            "parameters": {"bool": False},
        }
        httpclient.HTTPClient.post.assert_called_with("/executions", expected)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_runner_param_integer_conversion(self):
        self.shell.run(["run", "mockety.mock1", "int=30"])
        expected = {"action": "mockety.mock1", "user": None, "parameters": {"int": 30}}
        httpclient.HTTPClient.post.assert_called_with("/executions", expected)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_runner_param_float_conversion(self):
        self.shell.run(["run", "mockety.mock1", "float=3.01"])
        expected = {
            "action": "mockety.mock1",
            "user": None,
            "parameters": {"float": 3.01},
        }
        httpclient.HTTPClient.post.assert_called_with("/executions", expected)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_runner_param_json_conversion(self):
        self.shell.run(["run", "mockety.mock1", 'json={"a":1}'])
        expected = {
            "action": "mockety.mock1",
            "user": None,
            "parameters": {"json": {"a": 1}},
        }
        httpclient.HTTPClient.post.assert_called_with("/executions", expected)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_runner_param_array_conversion(self):
        self.shell.run(["run", "mockety.mock1", "list=one,two,three"])
        expected = {
            "action": "mockety.mock1",
            "user": None,
            "parameters": {"list": ["one", "two", "three"]},
        }
        httpclient.HTTPClient.post.assert_called_with("/executions", expected)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_runner_param_array_object_conversion(self):
        self.shell.run(
            [
                "run",
                "mockety.mock1",
                'list=[{"foo":1, "ponies":"rainbows"},{"pluto":false, "earth":true}]',
            ]
        )
        expected = {
            "action": "mockety.mock1",
            "user": None,
            "parameters": {
                "list": [
                    {"foo": 1, "ponies": "rainbows"},
                    {"pluto": False, "earth": True},
                ]
            },
        }
        httpclient.HTTPClient.post.assert_called_with("/executions", expected)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_param_bool_conversion(self):
        self.shell.run(["run", "mockety.mock2", "bool=false"])
        expected = {
            "action": "mockety.mock2",
            "user": None,
            "parameters": {"bool": False},
        }
        httpclient.HTTPClient.post.assert_called_with("/executions", expected)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_param_integer_conversion(self):
        self.shell.run(["run", "mockety.mock2", "int=30"])
        expected = {"action": "mockety.mock2", "user": None, "parameters": {"int": 30}}
        httpclient.HTTPClient.post.assert_called_with("/executions", expected)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_param_float_conversion(self):
        self.shell.run(["run", "mockety.mock2", "float=3.01"])
        expected = {
            "action": "mockety.mock2",
            "user": None,
            "parameters": {"float": 3.01},
        }
        httpclient.HTTPClient.post.assert_called_with("/executions", expected)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_param_json_conversion(self):
        self.shell.run(["run", "mockety.mock2", 'json={"a":1}'])
        expected = {
            "action": "mockety.mock2",
            "user": None,
            "parameters": {"json": {"a": 1}},
        }
        httpclient.HTTPClient.post.assert_called_with("/executions", expected)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_param_array_conversion(self):
        self.shell.run(["run", "mockety.mock2", "list=one,two,three"])
        expected = {
            "action": "mockety.mock2",
            "user": None,
            "parameters": {"list": ["one", "two", "three"]},
        }
        httpclient.HTTPClient.post.assert_called_with("/executions", expected)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_param_array_conversion_single_element_str(self):
        self.shell.run(["run", "mockety.mock2", "list=one"])
        expected = {
            "action": "mockety.mock2",
            "user": None,
            "parameters": {"list": ["one"]},
        }
        httpclient.HTTPClient.post.assert_called_with("/executions", expected)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_param_array_conversion_single_element_int(self):
        self.shell.run(["run", "mockety.mock2", "list=1"])
        expected = {
            "action": "mockety.mock2",
            "user": None,
            "parameters": {"list": [1]},
        }
        httpclient.HTTPClient.post.assert_called_with("/executions", expected)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_param_array_object_conversion(self):
        self.shell.run(
            [
                "run",
                "mockety.mock2",
                'list=[{"foo":1, "ponies":"rainbows"},{"pluto":false, "earth":true}]',
            ]
        )
        expected = {
            "action": "mockety.mock2",
            "user": None,
            "parameters": {
                "list": [
                    {"foo": 1, "ponies": "rainbows"},
                    {"pluto": False, "earth": True},
                ]
            },
        }
        httpclient.HTTPClient.post.assert_called_with("/executions", expected)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_param_dict_conversion_flag(self):
        """Ensure that the automatic conversion to dict based on colons only occurs with the flag"""

        self.shell.run(
            ["run", "mockety.mock2", "list=key1:value1,key2:value2", "--auto-dict"]
        )
        expected = {
            "action": "mockety.mock2",
            "user": None,
            "parameters": {"list": [{"key1": "value1", "key2": "value2"}]},
        }
        httpclient.HTTPClient.post.assert_called_with("/executions", expected)

        self.shell.run(["run", "mockety.mock2", "list=key1:value1,key2:value2"])
        expected = {
            "action": "mockety.mock2",
            "user": None,
            "parameters": {"list": ["key1:value1", "key2:value2"]},
        }
        httpclient.HTTPClient.post.assert_called_with("/executions", expected)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_param_value_with_equal_sign(self):
        self.shell.run(["run", "mockety.mock2", "key=foo=bar&ponies=unicorns"])
        expected = {
            "action": "mockety.mock2",
            "user": None,
            "parameters": {"key": "foo=bar&ponies=unicorns"},
        }
        httpclient.HTTPClient.post.assert_called_with("/executions", expected)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "delete",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_cancel_single_execution(self):
        self.shell.run(["execution", "cancel", "123"])
        httpclient.HTTPClient.delete.assert_called_with("/executions/123")

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "delete",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_cancel_multiple_executions(self):
        self.shell.run(["execution", "cancel", "123", "456", "789"])
        calls = [
            mock.call("/executions/123"),
            mock.call("/executions/456"),
            mock.call("/executions/789"),
        ]
        httpclient.HTTPClient.delete.assert_has_calls(calls)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "put",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_pause_single_execution(self):
        self.shell.run(["execution", "pause", "123"])
        expected = {"status": "pausing"}
        httpclient.HTTPClient.put.assert_called_with("/executions/123", expected)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "put",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_pause_multiple_executions(self):
        self.shell.run(["execution", "pause", "123", "456", "789"])
        expected = {"status": "pausing"}
        calls = [
            mock.call("/executions/123", expected),
            mock.call("/executions/456", expected),
            mock.call("/executions/789", expected),
        ]
        httpclient.HTTPClient.put.assert_has_calls(calls)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "put",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_resume_single_execution(self):
        self.shell.run(["execution", "resume", "123"])
        expected = {"status": "resuming"}
        httpclient.HTTPClient.put.assert_called_with("/executions/123", expected)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        models.ResourceManager, "get_by_name", mock.MagicMock(side_effect=get_by_name)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "put",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(LIVE_ACTION), 200, "OK")
        ),
    )
    def test_resume_multiple_executions(self):
        self.shell.run(["execution", "resume", "123", "456", "789"])
        expected = {"status": "resuming"}
        calls = [
            mock.call("/executions/123", expected),
            mock.call("/executions/456", expected),
            mock.call("/executions/789", expected),
        ]
        httpclient.HTTPClient.put.assert_has_calls(calls)



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import json
import logging
from pipes import quote as pquote

import requests


LOG = logging.getLogger(__name__)


def add_ssl_verify_to_kwargs(func):
    def decorate(*args, **kwargs):
        if isinstance(args[0], HTTPClient) and "https" in getattr(args[0], "root", ""):
            cacert = getattr(args[0], "cacert", None)
            kwargs["verify"] = cacert if cacert is not None else False
        return func(*args, **kwargs)

    return decorate


def add_basic_auth_creds_to_kwargs(func):
    """
    Add "auth" tuple parameter to the kwargs object which is passed to requests method in case it's
    present on the HTTPClient object instance.
    """

    def decorate(*args, **kwargs):
        # NOTE: When logging in using /v1/auth/tokens API endpoint, "auth" argument will already be
        # present since basic authentication is used to authenticate against auth service to obtain
        # a token.
        #
        # In such scenarios, we don't pass additional basic auth headers to the server.
        #
        # This is not ideal, because it means if additional proxy based http auth is enabled, user
        # may be able to authenticate against StackStorm auth service and obtain a valid auth token
        # without using additional basic auth credentials, but all the request of the API operations
        # on StackStorm API won't work without additional basic auth credentials.
        if (
            "auth" not in kwargs
            and isinstance(args[0], HTTPClient)
            and getattr(args[0], "basic_auth", None)
        ):
            kwargs["auth"] = args[0].basic_auth
        return func(*args, **kwargs)

    return decorate


def add_auth_token_to_headers(func):
    def decorate(*args, **kwargs):
        headers = kwargs.get("headers", dict())

        token = kwargs.pop("token", None)
        if token:
            headers["X-Auth-Token"] = str(token)
            kwargs["headers"] = headers

        api_key = kwargs.pop("api_key", None)
        if api_key:
            headers["St2-Api-Key"] = str(api_key)
            kwargs["headers"] = headers

        return func(*args, **kwargs)

    return decorate


def add_json_content_type_to_headers(func):
    def decorate(*args, **kwargs):
        headers = kwargs.get("headers", dict())
        content_type = headers.get("content-type", "application/json")
        headers["content-type"] = content_type
        kwargs["headers"] = headers
        return func(*args, **kwargs)

    return decorate


def get_url_without_trailing_slash(value):
    """
    Function which strips a trailing slash from the provided url if one is present.

    :param value: URL to format.
    :type value: ``str``

    :rtype: ``str``
    """
    result = value[:-1] if value.endswith("/") else value
    return result


class HTTPClient(object):
    def __init__(self, root, cacert=None, debug=False, basic_auth=None):
        self.root = get_url_without_trailing_slash(root)
        self.cacert = cacert
        self.debug = debug
        self.basic_auth = basic_auth

    @add_ssl_verify_to_kwargs
    @add_auth_token_to_headers
    @add_basic_auth_creds_to_kwargs
    def get(self, url, **kwargs):
        response = requests.get(self.root + url, **kwargs)
        response = self._response_hook(response=response)
        return response

    @add_ssl_verify_to_kwargs
    @add_auth_token_to_headers
    @add_json_content_type_to_headers
    @add_basic_auth_creds_to_kwargs
    def post(self, url, data, **kwargs):
        response = requests.post(self.root + url, json.dumps(data), **kwargs)
        response = self._response_hook(response=response)
        return response

    @add_ssl_verify_to_kwargs
    @add_auth_token_to_headers
    @add_basic_auth_creds_to_kwargs
    def post_raw(self, url, data, **kwargs):
        response = requests.post(self.root + url, data, **kwargs)
        response = self._response_hook(response=response)
        return response

    @add_ssl_verify_to_kwargs
    @add_auth_token_to_headers
    @add_json_content_type_to_headers
    @add_basic_auth_creds_to_kwargs
    def put(self, url, data, **kwargs):
        response = requests.put(self.root + url, json.dumps(data), **kwargs)
        response = self._response_hook(response=response)
        return response

    @add_ssl_verify_to_kwargs
    @add_auth_token_to_headers
    @add_json_content_type_to_headers
    @add_basic_auth_creds_to_kwargs
    def patch(self, url, data, **kwargs):
        response = requests.patch(self.root + url, data, **kwargs)
        response = self._response_hook(response=response)
        return response

    @add_ssl_verify_to_kwargs
    @add_auth_token_to_headers
    @add_basic_auth_creds_to_kwargs
    def delete(self, url, **kwargs):
        response = requests.delete(self.root + url, **kwargs)
        response = self._response_hook(response=response)
        return response

    def _response_hook(self, response):
        if self.debug:
            # Log cURL request line
            curl_line = self._get_curl_line_for_request(request=response.request)
            print("# -------- begin %d request ----------" % id(self))
            print(curl_line)
            print("# -------- begin %d response ----------" % (id(self)))
            print(response.text)
            print("# -------- end %d response ------------" % (id(self)))
            print("")

        return response

    def _get_curl_line_for_request(self, request):
        parts = ["curl"]

        # method
        method = request.method.upper()
        if method in ["HEAD"]:
            parts.extend(["--head"])
        else:
            parts.extend(["-X", pquote(method)])

        # headers
        for key, value in request.headers.items():
            parts.extend(["-H ", pquote("%s: %s" % (key, value))])

        # body
        if request.body:
            parts.extend(["--data-binary", pquote(request.body)])

        # URL
        parts.extend([pquote(request.url)])

        curl_line = " ".join(parts)
        return curl_line



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys


TYPE_TABLE = {
    dict: "object",
    list: "array",
    int: "integer",
    str: "string",
    float: "number",
    bool: "boolean",
    type(None): "null",
}

if sys.version_info[0] < 3:
    TYPE_TABLE[unicode] = "string"  # noqa  # pylint: disable=E0602


def _dict_to_schema(item):
    schema = {}
    for key, value in item.items():
        if isinstance(value, dict):
            schema[key] = {"type": "object", "parameters": _dict_to_schema(value)}
        else:
            schema[key] = {"type": TYPE_TABLE[type(value)]}

    return schema


def render_output_schema_from_output(output):
    """Given an action output produce a reasonable schema to match."""
    return _dict_to_schema(output)



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from st2client.commands import resource
from st2client.formatters import table
from st2client.models.reactor import TriggerInstance
from st2client.utils.date import format_isodate_for_user_timezone


class TriggerInstanceResendCommand(resource.ResourceCommand):
    def __init__(self, resource, *args, **kwargs):

        super(TriggerInstanceResendCommand, self).__init__(
            resource,
            kwargs.pop("name", "re-emit"),
            "Re-emit a particular trigger instance.",
            *args,
            **kwargs,
        )

        self.parser.add_argument(
            "id", nargs="?", metavar="id", help="ID of trigger instance to re-emit."
        )
        self.parser.add_argument(
            "-h",
            "--help",
            action="store_true",
            dest="help",
            help="Print usage for the given command.",
        )

    def run(self, args, **kwargs):
        return self.manager.re_emit(args.id)

    @resource.add_auth_token_to_kwargs_from_cli
    def run_and_print(self, args, **kwargs):
        ret = self.run(args, **kwargs)
        if "message" in ret:
            print(ret["message"])


class TriggerInstanceBranch(resource.ResourceBranch):
    def __init__(self, description, app, subparsers, parent_parser=None):
        super(TriggerInstanceBranch, self).__init__(
            TriggerInstance,
            description,
            app,
            subparsers,
            parent_parser=parent_parser,
            read_only=True,
            commands={
                "list": TriggerInstanceListCommand,
                "get": TriggerInstanceGetCommand,
            },
        )

        self.commands["re-emit"] = TriggerInstanceResendCommand(
            self.resource, self.app, self.subparsers, add_help=False
        )


class TriggerInstanceListCommand(resource.ResourceViewCommand):
    display_attributes = ["id", "trigger", "occurrence_time", "status"]

    attribute_transform_functions = {
        "occurrence_time": format_isodate_for_user_timezone
    }

    def __init__(self, resource, *args, **kwargs):

        self.default_limit = 50

        super(TriggerInstanceListCommand, self).__init__(
            resource,
            "list",
            "Get the list of the %s most recent %s."
            % (self.default_limit, resource.get_plural_display_name().lower()),
            *args,
            **kwargs,
        )

        self.resource_name = resource.get_plural_display_name().lower()
        self.group = self.parser.add_argument_group()
        self.parser.add_argument(
            "-n",
            "--last",
            type=int,
            dest="last",
            default=self.default_limit,
            help=(
                "List N most recent %s. Use -n -1 to fetch the full result \
                                       set."
                % self.resource_name
            ),
        )

        # Filter options
        self.group.add_argument(
            "--trigger", help="Trigger reference to filter the list."
        )

        self.parser.add_argument(
            "-tg",
            "--timestamp-gt",
            type=str,
            dest="timestamp_gt",
            default=None,
            help=(
                "Only return trigger instances with occurrence_time "
                "greater than the one provided. "
                "Use time in the format 2000-01-01T12:00:00.000Z"
            ),
        )
        self.parser.add_argument(
            "-tl",
            "--timestamp-lt",
            type=str,
            dest="timestamp_lt",
            default=None,
            help=(
                "Only return trigger instances with timestamp "
                "lower than the one provided. "
                "Use time in the format 2000-01-01T12:00:00.000Z"
            ),
        )

        self.group.add_argument(
            "--status",
            help="Can be pending, processing, processed or processing_failed.",
        )

        # Display options
        self.parser.add_argument(
            "-a",
            "--attr",
            nargs="+",
            default=self.display_attributes,
            help=(
                "List of attributes to include in the "
                'output. "all" will return all '
                "attributes."
            ),
        )
        self.parser.add_argument(
            "-w",
            "--width",
            nargs="+",
            type=int,
            default=None,
            help=("Set the width of columns in output."),
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        # Filtering options
        if args.trigger:
            kwargs["trigger"] = args.trigger
        if args.timestamp_gt:
            kwargs["timestamp_gt"] = args.timestamp_gt
        if args.timestamp_lt:
            kwargs["timestamp_lt"] = args.timestamp_lt
        if args.status:
            kwargs["status"] = args.status

        include_attributes = self._get_include_attributes(args=args)
        if include_attributes:
            include_attributes = ",".join(include_attributes)
            kwargs["params"] = {"include_attributes": include_attributes}

        return self.manager.query_with_count(limit=args.last, **kwargs)

    def run_and_print(self, args, **kwargs):
        instances, count = self.run(args, **kwargs)
        if args.json or args.yaml:
            self.print_output(
                reversed(instances),
                table.MultiColumnTable,
                attributes=args.attr,
                widths=args.width,
                json=args.json,
                yaml=args.yaml,
                attribute_transform_functions=self.attribute_transform_functions,
            )
        else:
            self.print_output(
                reversed(instances),
                table.MultiColumnTable,
                attributes=args.attr,
                widths=args.width,
                attribute_transform_functions=self.attribute_transform_functions,
            )
            if args.last and count and count > args.last:
                table.SingleRowTable.note_box(self.resource_name, args.last)


class TriggerInstanceGetCommand(resource.ResourceGetCommand):
    display_attributes = ["all"]
    attribute_display_order = ["id", "trigger", "occurrence_time", "payload"]

    pk_argument_name = "id"



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import getpass
import json
import logging
import os

import requests
import six
from six.moves.configparser import ConfigParser
from six.moves import http_client

from st2client.base import BaseCLIApp
from st2client import config_parser
from st2client.models.auth import ApiKey
from st2client.commands import resource
from st2client.commands.noop import NoopCommand
from st2client.exceptions.operations import OperationFailureException
from st2client.formatters import table


LOG = logging.getLogger(__name__)


class TokenCreateCommand(resource.ResourceCommand):

    display_attributes = ["user", "token", "expiry"]

    def __init__(self, resource, *args, **kwargs):

        kwargs["has_token_opt"] = False

        super(TokenCreateCommand, self).__init__(
            resource,
            kwargs.pop("name", "create"),
            "Authenticate user and acquire access token.",
            *args,
            **kwargs,
        )

        self.parser.add_argument("username", help="Name of the user to authenticate.")

        self.parser.add_argument(
            "-p",
            "--password",
            dest="password",
            help="Password for the user. If password is not provided, "
            "it will be prompted for.",
        )
        self.parser.add_argument(
            "-l",
            "--ttl",
            type=int,
            dest="ttl",
            default=None,
            help="The life span of the token in seconds. "
            "Max TTL configured by the admin supersedes this.",
        )
        self.parser.add_argument(
            "-t",
            "--only-token",
            action="store_true",
            dest="only_token",
            default=False,
            help="On successful authentication, print only token to the " "console.",
        )

    def run(self, args, **kwargs):
        if not args.password:
            args.password = getpass.getpass()
        instance = self.resource(ttl=args.ttl) if args.ttl else self.resource()
        return self.manager.create(
            instance, auth=(args.username, args.password), **kwargs
        )

    def run_and_print(self, args, **kwargs):
        instance = self.run(args, **kwargs)

        if args.only_token:
            print(instance.token)
        else:
            self.print_output(
                instance,
                table.PropertyValueTable,
                attributes=self.display_attributes,
                json=args.json,
                yaml=args.yaml,
            )


class LoginCommand(resource.ResourceCommand):
    display_attributes = ["user", "token", "expiry"]

    def __init__(self, resource, *args, **kwargs):

        kwargs["has_token_opt"] = False

        super(LoginCommand, self).__init__(
            resource,
            kwargs.pop("name", "create"),
            "Authenticate user, acquire access token, and update CLI config directory",
            *args,
            **kwargs,
        )

        self.parser.add_argument("username", help="Name of the user to authenticate.")

        self.parser.add_argument(
            "-p",
            "--password",
            dest="password",
            help="Password for the user. If password is not provided, "
            "it will be prompted for.",
        )
        self.parser.add_argument(
            "-l",
            "--ttl",
            type=int,
            dest="ttl",
            default=None,
            help="The life span of the token in seconds. "
            "Max TTL configured by the admin supersedes this.",
        )
        self.parser.add_argument(
            "-w",
            "--write-password",
            action="store_true",
            default=False,
            dest="write_password",
            help="Write the password in plain text to the config file "
            "(default is to omit it)",
        )

    def run(self, args, **kwargs):

        if not args.password:
            args.password = getpass.getpass()
        instance = self.resource(ttl=args.ttl) if args.ttl else self.resource()

        cli = BaseCLIApp()

        # Determine path to config file
        try:
            config_file = cli._get_config_file_path(args)
        except ValueError:
            # config file not found in args or in env, defaulting
            config_file = config_parser.ST2_CONFIG_PATH

        # Retrieve token
        manager = self.manager.create(
            instance, auth=(args.username, args.password), **kwargs
        )
        cli._cache_auth_token(token_obj=manager)

        # Update existing configuration with new credentials
        config = ConfigParser()
        config.read(config_file)

        # Modify config (and optionally populate with password)
        if not config.has_section("credentials"):
            config.add_section("credentials")

        config.set("credentials", "username", args.username)
        if args.write_password:
            config.set("credentials", "password", args.password)
        else:
            # Remove any existing password from config
            config.remove_option("credentials", "password")

        config_existed = os.path.exists(config_file)
        with open(config_file, "w") as cfg_file_out:
            config.write(cfg_file_out)
        # If we created the config file, correct the permissions
        if not config_existed:
            os.chmod(config_file, 0o660)

        return manager

    def run_and_print(self, args, **kwargs):
        try:
            self.run(args, **kwargs)
        except Exception as e:
            if self.app.client.debug:
                raise

            raise Exception(
                "Failed to log in as %s: %s" % (args.username, six.text_type(e))
            )

        print("Logged in as %s" % (args.username))

        if not args.write_password:
            # Note: Client can't depend and import from common so we need to hard-code this
            # default value
            token_expire_hours = 24

            print("")
            print(
                "Note: You didn't use --write-password option so the password hasn't been "
                "stored in the client config and you will need to login again in %s hours when "
                "the auth token expires." % (token_expire_hours)
            )
            print(
                'As an alternative, you can run st2 login command with the "--write-password" '
                "flag, but keep it mind this will cause it to store the password in plain-text "
                "in the client config file (~/.st2/config)."
            )


class WhoamiCommand(resource.ResourceCommand):
    display_attributes = ["user", "token", "expiry"]

    def __init__(self, resource, *args, **kwargs):

        kwargs["has_token_opt"] = False

        super(WhoamiCommand, self).__init__(
            resource,
            kwargs.pop("name", "create"),
            "Display the currently authenticated user",
            *args,
            **kwargs,
        )

    def run(self, args, **kwargs):
        user_info = self.app.client.get_user_info(**kwargs)
        return user_info

    def run_and_print(self, args, **kwargs):
        try:
            user_info = self.run(args, **kwargs)
        except Exception as e:
            response = getattr(e, "response", None)
            status_code = getattr(response, "status_code", None)
            is_unathorized_error = status_code == http_client.UNAUTHORIZED

            if response and is_unathorized_error:
                print("Not authenticated")
            else:
                print("Unable to retrieve currently logged-in user")

            if self.app.client.debug:
                raise

            return

        print('Currently logged in as "%s".' % (user_info["username"]))
        print("")
        print("Authentication method: %s" % (user_info["authentication"]["method"]))

        if user_info["authentication"]["method"] == "authentication token":
            print(
                "Authentication token expire time: %s"
                % (user_info["authentication"]["token_expire"])
            )

        print("")
        print("RBAC:")
        print(" - Enabled: %s" % (user_info["rbac"]["enabled"]))
        print(" - Roles: %s" % (", ".join(user_info["rbac"]["roles"])))


class ApiKeyBranch(resource.ResourceBranch):
    def __init__(self, description, app, subparsers, parent_parser=None):
        super(ApiKeyBranch, self).__init__(
            ApiKey,
            description,
            app,
            subparsers,
            parent_parser=parent_parser,
            commands={
                "list": ApiKeyListCommand,
                "get": ApiKeyGetCommand,
                "create": ApiKeyCreateCommand,
                "update": NoopCommand,
                "delete": ApiKeyDeleteCommand,
            },
        )

        self.commands["enable"] = ApiKeyEnableCommand(
            self.resource, self.app, self.subparsers
        )
        self.commands["disable"] = ApiKeyDisableCommand(
            self.resource, self.app, self.subparsers
        )
        self.commands["load"] = ApiKeyLoadCommand(
            self.resource, self.app, self.subparsers
        )


class ApiKeyListCommand(resource.ResourceListCommand):
    detail_display_attributes = ["all"]
    display_attributes = ["id", "user", "metadata"]

    def __init__(self, resource, *args, **kwargs):
        super(ApiKeyListCommand, self).__init__(resource, *args, **kwargs)

        self.parser.add_argument(
            "-u",
            "--user",
            type=str,
            help="Only return ApiKeys belonging to the provided user",
        )
        self.parser.add_argument(
            "-d", "--detail", action="store_true", help="Full list of attributes."
        )
        self.parser.add_argument(
            "--show-secrets", action="store_true", help="Full list of attributes."
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        filters = {}
        filters["user"] = args.user
        filters.update(**kwargs)
        # show_secrets is not a filter but a query param. There is some special
        # handling for filters in the get method which reuqires this odd hack.
        if args.show_secrets:
            params = filters.get("params", {})
            params["show_secrets"] = True
            filters["params"] = params
        return self.manager.get_all(**filters)

    def run_and_print(self, args, **kwargs):
        instances = self.run(args, **kwargs)
        attr = self.detail_display_attributes if args.detail else args.attr
        self.print_output(
            instances,
            table.MultiColumnTable,
            attributes=attr,
            widths=args.width,
            json=args.json,
            yaml=args.yaml,
        )


class ApiKeyGetCommand(resource.ResourceGetCommand):
    display_attributes = ["all"]
    attribute_display_order = ["id", "user", "metadata"]

    pk_argument_name = "key_or_id"  # name of the attribute which stores resource PK


class ApiKeyCreateCommand(resource.ResourceCommand):
    def __init__(self, resource, *args, **kwargs):
        super(ApiKeyCreateCommand, self).__init__(
            resource,
            "create",
            "Create a new %s." % resource.get_display_name().lower(),
            *args,
            **kwargs,
        )

        self.parser.add_argument(
            "-u",
            "--user",
            type=str,
            help="User for which to create API Keys.",
            default="",
        )
        self.parser.add_argument(
            "-m",
            "--metadata",
            type=json.loads,
            help="Optional metadata to associate with the API Keys.",
            default={},
        )
        self.parser.add_argument(
            "-k",
            "--only-key",
            action="store_true",
            dest="only_key",
            default=False,
            help="Only print API Key to the console on creation.",
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        data = {}
        if args.user:
            data["user"] = args.user
        if args.metadata:
            data["metadata"] = args.metadata
        instance = self.resource.deserialize(data)
        return self.manager.create(instance, **kwargs)

    def run_and_print(self, args, **kwargs):
        try:
            instance = self.run(args, **kwargs)
            if not instance:
                raise Exception("Server did not create instance.")
        except Exception as e:
            message = six.text_type(e)
            print("ERROR: %s" % (message))
            raise OperationFailureException(message)
        if args.only_key:
            print(instance.key)
        else:
            self.print_output(
                instance,
                table.PropertyValueTable,
                attributes=["all"],
                json=args.json,
                yaml=args.yaml,
            )


class ApiKeyLoadCommand(resource.ResourceCommand):
    def __init__(self, resource, *args, **kwargs):
        super(ApiKeyLoadCommand, self).__init__(
            resource,
            "load",
            "Load %s from a file." % resource.get_display_name().lower(),
            *args,
            **kwargs,
        )

        self.parser.add_argument(
            "file",
            help=(
                "JSON/YAML file containing the %s(s) to load."
                % resource.get_display_name().lower()
            ),
            default="",
        )

        self.parser.add_argument(
            "-w",
            "--width",
            nargs="+",
            type=int,
            default=None,
            help=("Set the width of columns in output."),
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        resources = resource.load_meta_file(args.file)
        if not resources:
            print(
                "No %s found in %s."
                % (self.resource.get_display_name().lower(), args.file)
            )
            return None
        if not isinstance(resources, list):
            resources = [resources]
        instances = []
        for res in resources:
            # pick only the meaningful properties.
            data = {
                "user": res["user"],  # required
                "key_hash": res["key_hash"],  # required
                "metadata": res.get("metadata", {}),
                "enabled": res.get("enabled", False),
            }

            if "id" in res:
                data["id"] = res["id"]

            instance = self.resource.deserialize(data)

            try:
                result = self.manager.update(instance, **kwargs)
            except requests.exceptions.HTTPError as e:
                if e.response.status_code == http_client.NOT_FOUND:
                    instance = self.resource.deserialize(data)
                    # Key doesn't exist yet, create it instead
                    result = self.manager.create(instance, **kwargs)
                else:
                    raise e

            instances.append(result)
        return instances

    def run_and_print(self, args, **kwargs):
        instances = self.run(args, **kwargs)
        if instances:
            self.print_output(
                instances,
                table.MultiColumnTable,
                attributes=ApiKeyListCommand.display_attributes,
                widths=args.width,
                json=args.json,
                yaml=args.yaml,
            )


class ApiKeyDeleteCommand(resource.ResourceDeleteCommand):
    pk_argument_name = "key_or_id"  # name of the attribute which stores resource PK


class ApiKeyEnableCommand(resource.ResourceEnableCommand):
    pk_argument_name = "key_or_id"  # name of the attribute which stores resource PK


class ApiKeyDisableCommand(resource.ResourceDisableCommand):
    pk_argument_name = "key_or_id"  # name of the attribute which stores resource PK



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import os

import unittest2
import mock

from st2client.utils.terminal import DEFAULT_TERMINAL_SIZE_COLUMNS
from st2client.utils.terminal import get_terminal_size_columns

__all__ = ["TerminalUtilsTestCase"]


class TerminalUtilsTestCase(unittest2.TestCase):
    def setUp(self):
        super(TerminalUtilsTestCase, self).setUp()

        if "COLUMNS" in os.environ:
            del os.environ["COLUMNS"]

    @mock.patch.dict(os.environ, {"LINES": "111", "COLUMNS": "222"})
    def test_get_terminal_size_columns_columns_environment_variable_has_precedence(
        self,
    ):
        # Verify that COLUMNS environment variables has precedence over other approaches
        columns = get_terminal_size_columns()

        self.assertEqual(columns, 222)

    # make sure that os.environ['COLUMNS'] isn't set so it can't override/screw-up this test
    @mock.patch.dict(os.environ, {})
    @mock.patch("fcntl.ioctl", mock.Mock(return_value="dummy"))
    @mock.patch("struct.unpack", mock.Mock(return_value=(333, 444)))
    def test_get_terminal_size_columns_stdout_is_used(self):
        columns = get_terminal_size_columns()
        self.assertEqual(columns, 444)

    @mock.patch("struct.unpack", mock.Mock(side_effect=Exception("a")))
    @mock.patch("subprocess.Popen")
    def test_get_terminal_size_subprocess_popen_is_used(self, mock_popen):
        mock_communicate = mock.Mock(return_value=["555 666"])

        mock_process = mock.Mock()
        mock_process.returncode = 0
        mock_process.communicate = mock_communicate

        mock_popen.return_value = mock_process

        columns = get_terminal_size_columns()
        self.assertEqual(columns, 666)

    @mock.patch("struct.unpack", mock.Mock(side_effect=Exception("a")))
    @mock.patch("subprocess.Popen", mock.Mock(side_effect=Exception("b")))
    def test_get_terminal_size_default_values_are_used(self):
        columns = get_terminal_size_columns()

        self.assertEqual(columns, DEFAULT_TERMINAL_SIZE_COLUMNS)



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import re

import six
import jsonschema
from jsonschema import Draft3Validator
from prompt_toolkit import prompt
from prompt_toolkit import token
from prompt_toolkit import validation

from st2client.exceptions.operations import OperationFailureException
from six.moves import range


POSITIVE_BOOLEAN = {"1", "y", "yes", "true"}
NEGATIVE_BOOLEAN = {"0", "n", "no", "nope", "nah", "false"}


class ReaderNotImplemented(OperationFailureException):
    pass


class DialogInterrupted(OperationFailureException):
    pass


class MuxValidator(validation.Validator):
    def __init__(self, validators, spec):
        super(MuxValidator, self).__init__()

        self.validators = validators
        self.spec = spec

    def validate(self, document):
        input = document.text

        for validator in self.validators:
            validator(input, self.spec)


class StringReader(object):
    def __init__(self, name, spec, prefix=None, secret=False, **kw):
        self.name = name
        self.spec = spec
        self.prefix = prefix or ""
        self.options = {"is_password": secret}

        self._construct_description()
        self._construct_template()
        self._construct_validators()

        self.options.update(kw)

    @staticmethod
    def condition(spec):
        return True

    @staticmethod
    def validate(input, spec):
        try:
            jsonschema.validate(input, spec, Draft3Validator)
        except jsonschema.ValidationError as e:
            raise validation.ValidationError(len(input), six.text_type(e))

    def read(self):
        message = self.template.format(self.prefix + self.name, **self.spec)
        response = prompt(message, **self.options)

        result = self.spec.get("default", None)

        if response:
            result = self._transform_response(response)

        return result

    def _construct_description(self):
        if "description" in self.spec:

            def get_bottom_toolbar_tokens(cli):
                return [(token.Token.Toolbar, self.spec["description"])]

            self.options["get_bottom_toolbar_tokens"] = get_bottom_toolbar_tokens

    def _construct_template(self):
        self.template = "{0}: "

        if "default" in self.spec:
            self.template = "{0} [{default}]: "

    def _construct_validators(self):
        self.options["validator"] = MuxValidator([self.validate], self.spec)

    def _transform_response(self, response):
        return response


class BooleanReader(StringReader):
    @staticmethod
    def condition(spec):
        return spec.get("type", None) == "boolean"

    @staticmethod
    def validate(input, spec):
        if not input and (not spec.get("required", None) or spec.get("default", None)):
            return

        if input.lower() not in POSITIVE_BOOLEAN | NEGATIVE_BOOLEAN:
            raise validation.ValidationError(
                len(input),
                "Does not look like boolean. Pick from [%s]"
                % ", ".join(POSITIVE_BOOLEAN | NEGATIVE_BOOLEAN),
            )

    def _construct_template(self):
        self.template = "{0} (boolean)"

        if "default" in self.spec:
            self.template += " [{}]: ".format(self.spec.get("default") and "y" or "n")
        else:
            self.template += ": "

    def _transform_response(self, response):
        if response.lower() in POSITIVE_BOOLEAN:
            return True
        if response.lower() in NEGATIVE_BOOLEAN:
            return False

        # Hopefully, it will never happen
        raise OperationFailureException(
            "Response neither positive no negative. "
            "Value have not been properly validated."
        )


class NumberReader(StringReader):
    @staticmethod
    def condition(spec):
        return spec.get("type", None) == "number"

    @staticmethod
    def validate(input, spec):
        if input:
            try:
                input = float(input)
            except ValueError as e:
                raise validation.ValidationError(len(input), six.text_type(e))

            super(NumberReader, NumberReader).validate(input, spec)

    def _construct_template(self):
        self.template = "{0} (float)"

        if "default" in self.spec:
            self.template += " [{default}]: ".format(default=self.spec.get("default"))
        else:
            self.template += ": "

    def _transform_response(self, response):
        return float(response)


class IntegerReader(StringReader):
    @staticmethod
    def condition(spec):
        return spec.get("type", None) == "integer"

    @staticmethod
    def validate(input, spec):
        if input:
            try:
                input = int(input)
            except ValueError as e:
                raise validation.ValidationError(len(input), six.text_type(e))

            super(IntegerReader, IntegerReader).validate(input, spec)

    def _construct_template(self):
        self.template = "{0} (integer)"

        if "default" in self.spec:
            self.template += " [{default}]: ".format(default=self.spec.get("default"))
        else:
            self.template += ": "

    def _transform_response(self, response):
        return int(response)


class SecretStringReader(StringReader):
    def __init__(self, *args, **kwargs):
        super(SecretStringReader, self).__init__(*args, secret=True, **kwargs)

    @staticmethod
    def condition(spec):
        return spec.get("secret", None)

    def _construct_template(self):
        self.template = "{0} (secret)"

        if "default" in self.spec:
            self.template += " [{default}]: ".format(default=self.spec.get("default"))
        else:
            self.template += ": "


class EnumReader(StringReader):
    @staticmethod
    def condition(spec):
        return spec.get("enum", None)

    @staticmethod
    def validate(input, spec):
        if not input and (not spec.get("required", None) or spec.get("default", None)):
            return

        if not input.isdigit():
            raise validation.ValidationError(len(input), "Not a number")

        enum = spec.get("enum")
        try:
            enum[int(input)]
        except IndexError:
            raise validation.ValidationError(len(input), "Out of bounds")

    def _construct_template(self):
        self.template = "{0}: "

        enum = self.spec.get("enum")
        for index, value in enumerate(enum):
            self.template += "\n {} - {}".format(index, value)

        num_options = len(enum)
        more = ""
        if num_options > 3:
            num_options = 3
            more = "..."
        options = [str(i) for i in range(0, num_options)]
        self.template += "\nChoose from {}{}".format(", ".join(options), more)

        if "default" in self.spec:
            self.template += " [{}]: ".format(enum.index(self.spec.get("default")))
        else:
            self.template += ": "

    def _transform_response(self, response):
        return self.spec.get("enum")[int(response)]


class ObjectReader(StringReader):
    @staticmethod
    def condition(spec):
        return spec.get("type", None) == "object"

    def read(self):
        prefix = "{}.".format(self.name)

        result = InteractiveForm(
            self.spec.get("properties", {}), prefix=prefix, reraise=True
        ).initiate_dialog()

        return result


class ArrayReader(StringReader):
    @staticmethod
    def condition(spec):
        return spec.get("type", None) == "array"

    @staticmethod
    def validate(input, spec):
        if not input and (not spec.get("required", None) or spec.get("default", None)):
            return

        for m in re.finditer(r"[^, ]+", input):
            index, item = m.start(), m.group()
            try:
                StringReader.validate(item, spec.get("items", {}))
            except validation.ValidationError as e:
                raise validation.ValidationError(index, six.text_type(e))

    def read(self):
        item_type = self.spec.get("items", {}).get("type", "string")

        if item_type not in ["string", "integer", "number", "boolean"]:
            message = (
                "Interactive mode does not support arrays of %s type yet" % item_type
            )
            raise ReaderNotImplemented(message)

        result = super(ArrayReader, self).read()

        return result

    def _construct_template(self):
        self.template = "{0} (comma-separated list)"

        if "default" in self.spec:
            self.template += " [{default}]: ".format(
                default=",".join(self.spec.get("default"))
            )
        else:
            self.template += ": "

    def _transform_response(self, response):
        return [item.strip() for item in response.split(",")]


class ArrayObjectReader(StringReader):
    @staticmethod
    def condition(spec):
        return (
            spec.get("type", None) == "array"
            and spec.get("items", {}).get("type") == "object"
        )

    def read(self):
        results = []
        properties = self.spec.get("items", {}).get("properties", {})
        message = (
            '~~~ Would you like to add another item to  "%s" array / list?' % self.name
        )

        is_continue = True
        index = 0
        while is_continue:
            prefix = "{name}[{index}].".format(name=self.name, index=index)
            results.append(
                InteractiveForm(
                    properties, prefix=prefix, reraise=True
                ).initiate_dialog()
            )

            index += 1
            if Question(message, {"default": "y"}).read() != "y":
                is_continue = False

        return results


class ArrayEnumReader(EnumReader):
    def __init__(self, name, spec, prefix=None):
        self.items = spec.get("items", {})

        super(ArrayEnumReader, self).__init__(name, spec, prefix)

    @staticmethod
    def condition(spec):
        return spec.get("type", None) == "array" and "enum" in spec.get("items", {})

    @staticmethod
    def validate(input, spec):
        if not input and (not spec.get("required", None) or spec.get("default", None)):
            return

        for m in re.finditer(r"[^, ]+", input):
            index, item = m.start(), m.group()
            try:
                EnumReader.validate(item, spec.get("items", {}))
            except validation.ValidationError as e:
                raise validation.ValidationError(index, six.text_type(e))

    def _construct_template(self):
        self.template = "{0}: "

        enum = self.items.get("enum")
        for index, value in enumerate(enum):
            self.template += "\n {} - {}".format(index, value)

        num_options = len(enum)
        more = ""
        if num_options > 3:
            num_options = 3
            more = "..."
        options = [str(i) for i in range(0, num_options)]
        self.template += "\nChoose from {}{}".format(", ".join(options), more)

        if "default" in self.spec:
            default_choises = [
                str(enum.index(item)) for item in self.spec.get("default")
            ]
            self.template += " [{}]: ".format(", ".join(default_choises))
        else:
            self.template += ": "

    def _transform_response(self, response):
        result = []

        for i in (item.strip() for item in response.split(",")):
            if i:
                result.append(self.items.get("enum")[int(i)])

        return result


class InteractiveForm(object):
    readers = [
        EnumReader,
        BooleanReader,
        NumberReader,
        IntegerReader,
        ObjectReader,
        ArrayEnumReader,
        ArrayObjectReader,
        ArrayReader,
        SecretStringReader,
        StringReader,
    ]

    def __init__(self, schema, prefix=None, reraise=False):
        self.schema = schema
        self.prefix = prefix
        self.reraise = reraise

    def initiate_dialog(self):
        result = {}

        try:
            for field in self.schema:
                try:
                    result[field] = self._read_field(field)
                except ReaderNotImplemented as e:
                    print("%s. Skipping..." % six.text_type(e))
        except DialogInterrupted:
            if self.reraise:
                raise
            print("Dialog interrupted.")

        return result

    def _read_field(self, field):
        spec = self.schema[field]

        reader = None

        for Reader in self.readers:
            if Reader.condition(spec):
                reader = Reader(field, spec, prefix=self.prefix)
                break

        if not reader:
            raise ReaderNotImplemented("No reader for the field spec")

        try:
            return reader.read()
        except KeyboardInterrupt:
            raise DialogInterrupted()


class Question(StringReader):
    def __init__(self, message, spec=None):
        if not spec:
            spec = {}

        super(Question, self).__init__(message, spec)



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from st2client.models import core
from st2client.models.action import Execution
from st2client.models.action_alias import ActionAlias
from st2client.models.action_alias import ActionAliasMatch
from st2client.commands import resource
from st2client.commands.action import ActionRunCommandMixin
from st2client.formatters import table


__all__ = ["ActionAliasBranch", "ActionAliasMatchCommand", "ActionAliasExecuteCommand"]


class ActionAliasBranch(resource.ResourceBranch):
    def __init__(self, description, app, subparsers, parent_parser=None):
        super(ActionAliasBranch, self).__init__(
            ActionAlias,
            description,
            app,
            subparsers,
            parent_parser=parent_parser,
            read_only=False,
            commands={"list": ActionAliasListCommand, "get": ActionAliasGetCommand},
        )

        self.commands["match"] = ActionAliasMatchCommand(
            self.resource, self.app, self.subparsers, add_help=True
        )
        self.commands["execute"] = ActionAliasExecuteCommand(
            self.resource, self.app, self.subparsers, add_help=True
        )
        self.commands["test"] = ActionAliasTestCommand(
            self.resource, self.app, self.subparsers, add_help=True
        )


class ActionAliasListCommand(resource.ContentPackResourceListCommand):
    display_attributes = ["ref", "pack", "description", "enabled"]


class ActionAliasGetCommand(resource.ContentPackResourceGetCommand):
    display_attributes = ["all"]
    attribute_display_order = [
        "id",
        "ref",
        "pack",
        "name",
        "description",
        "enabled",
        "action_ref",
        "formats",
    ]


class ActionAliasMatchCommand(resource.ResourceCommand):
    display_attributes = ["name", "description"]

    def __init__(self, resource, *args, **kwargs):
        super(ActionAliasMatchCommand, self).__init__(
            resource,
            "match",
            "Get the %s that match the command text."
            % resource.get_display_name().lower(),
            *args,
            **kwargs,
        )

        self.parser.add_argument(
            "match_text",
            metavar="command",
            help=(
                "Get the %s that match the command text."
                % resource.get_display_name().lower()
            ),
        )
        self.parser.add_argument(
            "-a",
            "--attr",
            nargs="+",
            default=self.display_attributes,
            help=(
                "List of attributes to include in the "
                'output. "all" will return all '
                "attributes."
            ),
        )
        self.parser.add_argument(
            "-w",
            "--width",
            nargs="+",
            type=int,
            default=None,
            help=("Set the width of columns in output."),
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        alias_match = ActionAliasMatch()
        alias_match.command = args.match_text

        match, _ = self.manager.match(alias_match, **kwargs)
        return [match]

    def run_and_print(self, args, **kwargs):
        instances = self.run(args, **kwargs)
        self.print_output(
            instances,
            table.MultiColumnTable,
            attributes=args.attr,
            widths=args.width,
            json=args.json,
            yaml=args.yaml,
        )


class ActionAliasExecuteCommand(resource.ResourceCommand):
    display_attributes = ["name"]

    def __init__(self, resource, *args, **kwargs):
        super(ActionAliasExecuteCommand, self).__init__(
            resource,
            "execute",
            (
                "Execute the command text by finding a matching %s."
                % resource.get_display_name().lower()
            ),
            *args,
            **kwargs,
        )

        self.parser.add_argument(
            "command_text",
            metavar="command",
            help=(
                "Execute the command text by finding a matching %s."
                % resource.get_display_name().lower()
            ),
        )
        self.parser.add_argument(
            "-u",
            "--user",
            type=str,
            default=None,
            help="User under which to run the action (admins only).",
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        payload = core.Resource()
        payload.command = args.command_text
        payload.user = args.user or ""
        payload.source_channel = "cli"

        alias_execution_mgr = self.app.client.managers["ActionAliasExecution"]
        execution = alias_execution_mgr.match_and_execute(payload)
        return execution

    def run_and_print(self, args, **kwargs):
        execution = self.run(args, **kwargs)
        print("Matching Action-alias: '%s'" % execution.actionalias["ref"])
        print(
            "To get the results, execute:\n st2 execution get %s"
            % (execution.execution["id"])
        )


class ActionAliasTestCommand(ActionRunCommandMixin, resource.ResourceCommand):
    display_attributes = ["name"]

    def __init__(self, resource, *args, **kwargs):
        super(ActionAliasTestCommand, self).__init__(
            resource,
            "test",
            (
                "Execute the command text by finding a matching %s and format the result."
                % resource.get_display_name().lower()
            ),
            *args,
            **kwargs,
        )

        self.parser.add_argument(
            "command_text",
            metavar="command",
            help=(
                "Execute the command text by finding a matching %s."
                % resource.get_display_name().lower()
            ),
        )
        self.parser.add_argument(
            "-u",
            "--user",
            type=str,
            default=None,
            help="User under which to run the action (admins only).",
        )

        self._add_common_options()
        self.parser.add_argument(
            "-a",
            "--async",
            action="store_true",
            dest="action_async",
            help="Do not wait for action to finish.",
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        payload = core.Resource()
        payload.command = args.command_text
        payload.user = args.user or ""
        payload.source_channel = "cli"

        alias_execution_mgr = self.app.client.managers["ActionAliasExecution"]
        execution = alias_execution_mgr.match_and_execute(payload)
        return execution

    def run_and_print(self, args, **kwargs):
        # 1. Trigger the execution via alias
        print("Triggering execution via action alias")
        print("")

        # NOTE: This will return an error and abort if command matches no aliases so no additional
        # checks are needed
        result = self.run(args, **kwargs)
        execution = Execution.deserialize(result.execution)

        # 2. Wait for it to complete
        print(
            "Execution (%s) has been started, waiting for it to finish..."
            % (execution.id)
        )
        print("")

        action_exec_mgr = self.app.client.managers["Execution"]
        execution = self._get_execution_result(
            execution=execution, action_exec_mgr=action_exec_mgr, args=args, **kwargs
        )
        execution_id = execution.id

        # 3. Run chatops.format_result action with the result of the completed execution
        print("")
        print(f"Execution ({execution_id}) has finished, rendering result...")
        print("")

        format_execution = Execution()
        format_execution.action = "chatops.format_execution_result"
        format_execution.parameters = {"execution_id": execution_id}
        format_execution.user = args.user or ""

        format_execution = action_exec_mgr.create(format_execution, **kwargs)

        print(
            "Execution (%s) has been started, waiting for it to finish..."
            % (format_execution.id)
        )
        print("")

        # 4. Wait for chatops.format_execution_result to finish and print the result
        format_execution = self._get_execution_result(
            execution=format_execution,
            action_exec_mgr=action_exec_mgr,
            args=args,
            force_retry_on_finish=True,
            **kwargs,
        )

        print("")
        print("Formatted ChatOps result message")
        print("")
        print("=" * 80)
        print(format_execution.result["result"]["message"])
        print("=" * 80)
        print("")



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from jsonpath_rw import parse
import re

# A simple expression is defined as a series of letters [a-zA-Z], numbers [0-9],
# dashes '-', and underscores '_' separated by one period '.'.
# A simple expression must not start with a period.
# A simple expression must not end with a period.
# A simple expression must not have more than one period in succession, ie. in
# between each period must be one or more of the valid non-period characters.
#
# Examples of valid "simple expressions":
#  abc
#  abc.def.ghi
#
# Examples of non-simple expressions:
#  .aaa
#  a..b
#  abc.
#  a(*
SIMPLE_EXPRESSION_REGEX = r"^([a-zA-Z0-9\-_]+\.)*([a-zA-Z0-9\-_]+)$"
SIMPLE_EXPRESSION_REGEX_CMPL = re.compile(SIMPLE_EXPRESSION_REGEX)


def _get_value_simple(doc, key):
    """
    Extracts a value from a nested set of dictionaries 'doc' based on
    a 'key' string.
    The key string is expected to be of the format 'x.y.z'
    where each component in the string is a key in a dictionary separated
    by '.' to denote the next key is in a nested dictionary.

    Returns the extracted value from the key specified (if found)
    Returns None if the key can not be found
    """
    split_key = key.split(".")
    if not split_key:
        return None

    value = doc
    for k in split_key:
        if isinstance(value, dict) and k in value:
            value = value[k]
        else:
            return None
    return value


def _get_value_complex(doc, key):
    """
    Extracts a value from a nested set of dictionaries 'doc' based on
    a 'key' string.
    The key is expected to be a jsonpath_rw expression:
    http://jsonpath-rw.readthedocs.io/en/stable/

    Returns the extracted value from the key specified (if found)
    Returns None if the key can not be found
    """
    jsonpath_expr = parse(key)
    matches = jsonpath_expr.find(doc)
    value = None if len(matches) < 1 else matches[0].value
    return value


def get_value(doc, key):
    if not key:
        raise ValueError("key is None or empty: '{}'".format(key))

    if not isinstance(doc, dict):
        raise ValueError(
            "doc is not an instance of dict: type={} value='{}'".format(type(doc), doc)
        )
    # jsonpath_rw can be very slow when processing expressions.
    # In the case of a simple expression we've created a "fast path" that avoids
    # the complexity introduced by running jsonpath_rw code.
    # For more complex expressions we fall back to using jsonpath_rw.
    # This provides flexibility and increases performance in the base case.
    match = SIMPLE_EXPRESSION_REGEX_CMPL.match(key)
    if match:
        return _get_value_simple(doc, key)
    else:
        return _get_value_complex(doc, key)


def get_kvps(doc, keys):
    """
    Extracts one or more keys ('keys' can be a string or list of strings)
    from the dictionary 'doc'.

    Return a subset of 'doc' with only the 'keys' specified as members, all
    other data in the dictionary will be filtered out.
    Return an empty dict if no keys are found.
    """
    if not isinstance(keys, list):
        keys = [keys]

    new_doc = {}
    for key in keys:
        value = get_value(doc, key)
        if value is not None:
            nested = new_doc
            while "." in key:
                attr = key[: key.index(".")]
                if attr not in nested:
                    nested[attr] = {}
                nested = nested[attr]
                key = key[key.index(".") + 1 :]
            nested[key] = value

    return new_doc



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging

from st2client.models import core


LOG = logging.getLogger(__name__)


class Timer(core.Resource):
    _alias = "Timer"
    _display_name = "Timer"
    _plural = "Timers"
    _plural_display_name = "Timers"



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import
import six


def unescape(s):
    """
    Action execution escapes escaped chars in result (i.e. \n is stored as \\n).
    This function unescapes those chars.
    """
    if isinstance(s, six.string_types):
        s = s.replace("\\n", "\n")
        s = s.replace("\\r", "\r")
        s = s.replace('\\"', '"')

    return s


def dedupe_newlines(s):
    """yaml.safe_dump converts single newlines to double.

    Since we're printing this output and not loading it, we should
    deduplicate them.
    """

    if isinstance(s, six.string_types):
        s = s.replace("\n\n", "\n")

    return s


def strip_carriage_returns(s):
    if isinstance(s, six.string_types):
        s = s.replace("\\r", "")
        s = s.replace("\r", "")

    return s



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging

from st2client.models.policy import Policy, PolicyType
from st2client.commands import resource


LOG = logging.getLogger(__name__)


class PolicyTypeBranch(resource.ResourceBranch):
    def __init__(self, description, app, subparsers, parent_parser=None):
        super(PolicyTypeBranch, self).__init__(
            PolicyType,
            description,
            app,
            subparsers,
            parent_parser=parent_parser,
            read_only=True,
            commands={"list": PolicyTypeListCommand, "get": PolicyTypeGetCommand},
        )


class PolicyTypeListCommand(resource.ResourceListCommand):
    display_attributes = ["id", "resource_type", "name", "description"]

    def __init__(self, resource, *args, **kwargs):
        super(PolicyTypeListCommand, self).__init__(resource, *args, **kwargs)

        self.parser.add_argument(
            "-r",
            "--resource-type",
            type=str,
            dest="resource_type",
            help="Return policy types for the resource type.",
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        if args.resource_type:
            filters = {"resource_type": args.resource_type}
            filters.update(**kwargs)
            instances = self.manager.query(**filters)
            return instances
        else:
            return self.manager.get_all(**kwargs)


class PolicyTypeGetCommand(resource.ResourceGetCommand):
    pk_argument_name = "ref_or_id"

    def get_resource(self, ref_or_id, **kwargs):
        return self.get_resource_by_ref_or_id(ref_or_id=ref_or_id, **kwargs)


class PolicyBranch(resource.ResourceBranch):
    def __init__(self, description, app, subparsers, parent_parser=None):
        super(PolicyBranch, self).__init__(
            Policy,
            description,
            app,
            subparsers,
            parent_parser=parent_parser,
            commands={
                "list": PolicyListCommand,
                "get": PolicyGetCommand,
                "update": PolicyUpdateCommand,
                "delete": PolicyDeleteCommand,
            },
        )


class PolicyListCommand(resource.ContentPackResourceListCommand):
    display_attributes = ["ref", "resource_ref", "policy_type", "enabled"]

    def __init__(self, resource, *args, **kwargs):
        super(PolicyListCommand, self).__init__(resource, *args, **kwargs)

        self.parser.add_argument(
            "-r",
            "--resource-ref",
            type=str,
            dest="resource_ref",
            help="Return policies for the resource ref.",
        )
        self.parser.add_argument(
            "-pt",
            "--policy-type",
            type=str,
            dest="policy_type",
            help="Return policies of the policy type.",
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        filters = {}
        if args.pack:
            filters["pack"] = args.pack
        if args.resource_ref:
            filters["resource_ref"] = args.resource_ref
        if args.policy_type:
            filters["policy_type"] = args.policy_type
        filters.update(**kwargs)
        include_attributes = self._get_include_attributes(args=args)
        if include_attributes:
            include_attributes = ",".join(include_attributes)
            filters["params"] = {"include_attributes": include_attributes}
        return self.manager.query(**filters)


class PolicyGetCommand(resource.ContentPackResourceGetCommand):
    display_attributes = ["all"]
    attribute_display_order = [
        "id",
        "ref",
        "pack",
        "name",
        "description",
        "enabled",
        "resource_ref",
        "policy_type",
        "parameters",
    ]


class PolicyUpdateCommand(resource.ContentPackResourceUpdateCommand):
    pass


class PolicyDeleteCommand(resource.ContentPackResourceDeleteCommand):
    pass



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import json
import logging
import mock
import unittest2

from st2client.utils import jsutil


LOG = logging.getLogger(__name__)

DOC = {
    "a01": 1,
    "b01": 2,
    "c01": {"c11": 3, "d12": 4, "c13": {"c21": 5, "c22": 6}, "c14": [7, 8, 9]},
}

DOC_IP_ADDRESS = {
    "ips": {
        "192.168.1.1": {"hostname": "router.domain.tld"},
        "192.168.1.10": {"hostname": "server.domain.tld"},
    }
}


class TestGetValue(unittest2.TestCase):
    def test_dot_notation(self):
        self.assertEqual(jsutil.get_value(DOC, "a01"), 1)
        self.assertEqual(jsutil.get_value(DOC, "c01.c11"), 3)
        self.assertEqual(jsutil.get_value(DOC, "c01.c13.c22"), 6)
        self.assertEqual(jsutil.get_value(DOC, "c01.c13"), {"c21": 5, "c22": 6})
        self.assertListEqual(jsutil.get_value(DOC, "c01.c14"), [7, 8, 9])

    def test_dot_notation_with_val_error(self):
        self.assertRaises(ValueError, jsutil.get_value, DOC, None)
        self.assertRaises(ValueError, jsutil.get_value, DOC, "")
        self.assertRaises(ValueError, jsutil.get_value, json.dumps(DOC), "a01")

    def test_dot_notation_with_key_error(self):
        self.assertIsNone(jsutil.get_value(DOC, "d01"))
        self.assertIsNone(jsutil.get_value(DOC, "a01.a11"))
        self.assertIsNone(jsutil.get_value(DOC, "c01.c11.c21.c31"))
        self.assertIsNone(jsutil.get_value(DOC, "c01.c14.c31"))

    def test_ip_address(self):
        self.assertEqual(
            jsutil.get_value(DOC_IP_ADDRESS, 'ips."192.168.1.1"'),
            {"hostname": "router.domain.tld"},
        )

    def test_chars_nums_dashes_underscores_calls_simple(self):
        for char in "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-_":
            with mock.patch("st2client.utils.jsutil._get_value_simple") as mock_simple:
                jsutil.get_value(DOC, char)
                mock_simple.assert_called_with(DOC, char)

    def test_symbols_calls_complex(self):
        for char in "`~!@#$%^&&*()=+{}[]|\\;:'\"<>,./?":
            with mock.patch(
                "st2client.utils.jsutil._get_value_complex"
            ) as mock_complex:
                jsutil.get_value(DOC, char)
                mock_complex.assert_called_with(DOC, char)

    @mock.patch("st2client.utils.jsutil._get_value_simple")
    def test_single_key_calls_simple(self, mock__get_value_simple):
        jsutil.get_value(DOC, "a01")
        mock__get_value_simple.assert_called_with(DOC, "a01")

    @mock.patch("st2client.utils.jsutil._get_value_simple")
    def test_dot_notation_calls_simple(self, mock__get_value_simple):
        jsutil.get_value(DOC, "c01.c11")
        mock__get_value_simple.assert_called_with(DOC, "c01.c11")

    @mock.patch("st2client.utils.jsutil._get_value_complex")
    def test_ip_address_calls_complex(self, mock__get_value_complex):
        jsutil.get_value(DOC_IP_ADDRESS, 'ips."192.168.1.1"')
        mock__get_value_complex.assert_called_with(DOC_IP_ADDRESS, 'ips."192.168.1.1"')

    @mock.patch("st2client.utils.jsutil._get_value_complex")
    def test_beginning_dot_calls_complex(self, mock__get_value_complex):
        jsutil.get_value(DOC, ".c01.c11")
        mock__get_value_complex.assert_called_with(DOC, ".c01.c11")

    @mock.patch("st2client.utils.jsutil._get_value_complex")
    def test_ending_dot_calls_complex(self, mock__get_value_complex):
        jsutil.get_value(DOC, "c01.c11.")
        mock__get_value_complex.assert_called_with(DOC, "c01.c11.")

    @mock.patch("st2client.utils.jsutil._get_value_complex")
    def test_double_dot_calls_complex(self, mock__get_value_complex):
        jsutil.get_value(DOC, "c01..c11")
        mock__get_value_complex.assert_called_with(DOC, "c01..c11")


class TestGetKeyValuePairs(unittest2.TestCase):
    def test_select_kvps(self):
        self.assertEqual(jsutil.get_kvps(DOC, ["a01"]), {"a01": 1})
        self.assertEqual(jsutil.get_kvps(DOC, ["c01.c11"]), {"c01": {"c11": 3}})
        self.assertEqual(
            jsutil.get_kvps(DOC, ["c01.c13.c22"]), {"c01": {"c13": {"c22": 6}}}
        )
        self.assertEqual(
            jsutil.get_kvps(DOC, ["c01.c13"]), {"c01": {"c13": {"c21": 5, "c22": 6}}}
        )
        self.assertEqual(jsutil.get_kvps(DOC, ["c01.c14"]), {"c01": {"c14": [7, 8, 9]}})
        self.assertEqual(
            jsutil.get_kvps(DOC, ["a01", "c01.c11", "c01.c13.c21"]),
            {"a01": 1, "c01": {"c11": 3, "c13": {"c21": 5}}},
        )
        self.assertEqual(
            jsutil.get_kvps(
                DOC_IP_ADDRESS, ['ips."192.168.1.1"', 'ips."192.168.1.10".hostname']
            ),
            {
                "ips": {
                    '"192': {
                        "168": {
                            "1": {
                                '1"': {"hostname": "router.domain.tld"},
                                '10"': {"hostname": "server.domain.tld"},
                            }
                        }
                    }
                }
            },
        )

    def test_select_kvps_with_val_error(self):
        self.assertRaises(ValueError, jsutil.get_kvps, DOC, [None])
        self.assertRaises(ValueError, jsutil.get_kvps, DOC, [""])
        self.assertRaises(ValueError, jsutil.get_kvps, json.dumps(DOC), ["a01"])

    def test_select_kvps_with_key_error(self):
        self.assertEqual(jsutil.get_kvps(DOC, ["d01"]), {})
        self.assertEqual(jsutil.get_kvps(DOC, ["a01.a11"]), {})
        self.assertEqual(jsutil.get_kvps(DOC, ["c01.c11.c21.c31"]), {})
        self.assertEqual(jsutil.get_kvps(DOC, ["c01.c14.c31"]), {})
        self.assertEqual(
            jsutil.get_kvps(DOC, ["a01", "c01.c11", "c01.c13.c23"]),
            {"a01": 1, "c01": {"c11": 3}},
        )



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import os

__all__ = ["DisplayColors", "format_status"]


TERMINAL_SUPPORTS_ANSI_CODES = [
    "xterm",
    "xterm-color",
    "screen",
    "vt100",
    "vt100-color",
    "xterm-256color",
]

DISABLED = os.environ.get("ST2_COLORIZE", "")


class DisplayColors(object):
    RED = "\033[91m"
    PURPLE = "\033[35m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    BLUE = "\033[94m"
    BROWN = "\033[33m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"

    @staticmethod
    def colorize(value, color=""):
        # TODO: use list of supported terminals
        term = os.environ.get("TERM", None)

        if term is None or term.lower() not in TERMINAL_SUPPORTS_ANSI_CODES:
            # Terminal doesn't support colors
            return value

        if DISABLED or not color:
            return value

        return "%s%s%s" % (color, value, DisplayColors.ENDC)


# Lookup table
STATUS_LOOKUP = {
    "succeeded": DisplayColors.GREEN,
    "delayed": DisplayColors.BLUE,
    "failed": DisplayColors.RED,
    "timeout": DisplayColors.BROWN,
    "running": DisplayColors.YELLOW,
}


def format_status(value):
    # Support status values with elapsed info
    split = value.split("(", 1)

    if len(split) == 2:
        status = split[0].strip()
        remainder = "(" + split[1]
    else:
        status = value
        remainder = ""

    color = STATUS_LOOKUP.get(status, DisplayColors.YELLOW)
    result = DisplayColors.colorize(status, color)

    if remainder:
        result = result + " " + remainder
    return result



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from st2client.models.core import Resource
from st2client.models.trace import Trace
from st2client.models.reactor import TriggerInstance, Rule
from st2client.models.action import Execution
from st2client.formatters import table
from st2client.formatters import execution as execution_formatter
from st2client.commands import resource
from st2client.utils.date import format_isodate_for_user_timezone


TRACE_ATTRIBUTE_DISPLAY_ORDER = [
    "id",
    "trace_tag",
    "action_executions",
    "rules",
    "trigger_instances",
    "start_timestamp",
]

TRACE_HEADER_DISPLAY_ORDER = ["id", "trace_tag", "start_timestamp"]

TRACE_COMPONENT_DISPLAY_LABELS = ["id", "type", "ref", "updated_at"]

TRACE_DISPLAY_ATTRIBUTES = ["all"]

TRIGGER_INSTANCE_DISPLAY_OPTIONS = [
    "all",
    "trigger-instances",
    "trigger_instances",
    "triggerinstances",
    "triggers",
]

ACTION_EXECUTION_DISPLAY_OPTIONS = [
    "all",
    "executions",
    "action-executions",
    "action_executions",
    "actionexecutions",
    "actions",
]


class TraceBranch(resource.ResourceBranch):
    def __init__(self, description, app, subparsers, parent_parser=None):
        super(TraceBranch, self).__init__(
            Trace,
            description,
            app,
            subparsers,
            parent_parser=parent_parser,
            read_only=True,
            commands={"list": TraceListCommand, "get": TraceGetCommand},
        )


class SingleTraceDisplayMixin(object):
    def print_trace_details(self, trace, args, **kwargs):
        options = {
            "attributes": TRACE_ATTRIBUTE_DISPLAY_ORDER
            if args.json
            else TRACE_HEADER_DISPLAY_ORDER
        }
        options["json"] = args.json
        options["yaml"] = args.yaml
        options["attribute_transform_functions"] = self.attribute_transform_functions

        formatter = execution_formatter.ExecutionResult

        self.print_output(trace, formatter, **options)

        # Everything should be printed if we are printing json.
        if args.json or args.yaml:
            return

        components = []
        if any(attr in args.attr for attr in TRIGGER_INSTANCE_DISPLAY_OPTIONS):
            components.extend(
                [
                    Resource(
                        **{
                            "id": trigger_instance["object_id"],
                            "type": TriggerInstance._alias.lower(),
                            "ref": trigger_instance["ref"],
                            "updated_at": trigger_instance["updated_at"],
                        }
                    )
                    for trigger_instance in trace.trigger_instances
                ]
            )
        if any(attr in args.attr for attr in ["all", "rules"]):
            components.extend(
                [
                    Resource(
                        **{
                            "id": rule["object_id"],
                            "type": Rule._alias.lower(),
                            "ref": rule["ref"],
                            "updated_at": rule["updated_at"],
                        }
                    )
                    for rule in trace.rules
                ]
            )
        if any(attr in args.attr for attr in ACTION_EXECUTION_DISPLAY_OPTIONS):
            components.extend(
                [
                    Resource(
                        **{
                            "id": execution["object_id"],
                            "type": Execution._alias.lower(),
                            "ref": execution["ref"],
                            "updated_at": execution["updated_at"],
                        }
                    )
                    for execution in trace.action_executions
                ]
            )
        if components:
            components.sort(key=lambda resource: resource.updated_at)
            self.print_output(
                components,
                table.MultiColumnTable,
                attributes=TRACE_COMPONENT_DISPLAY_LABELS,
                json=args.json,
                yaml=args.yaml,
            )


class TraceListCommand(resource.ResourceCommand, SingleTraceDisplayMixin):
    display_attributes = ["id", "uid", "trace_tag", "start_timestamp"]

    attribute_transform_functions = {
        "start_timestamp": format_isodate_for_user_timezone
    }

    attribute_display_order = TRACE_ATTRIBUTE_DISPLAY_ORDER

    def __init__(self, resource, *args, **kwargs):

        self.default_limit = 50

        super(TraceListCommand, self).__init__(
            resource,
            "list",
            "Get the list of the %s most recent %s."
            % (self.default_limit, resource.get_plural_display_name().lower()),
            *args,
            **kwargs,
        )

        self.resource_name = resource.get_plural_display_name().lower()
        self.group = self.parser.add_mutually_exclusive_group()
        self.parser.add_argument(
            "-n",
            "--last",
            type=int,
            dest="last",
            default=self.default_limit,
            help=(
                "List N most recent %s. Use -n -1 to fetch the full result \
                                       set."
                % self.resource_name
            ),
        )
        self.parser.add_argument(
            "-s",
            "--sort",
            type=str,
            dest="sort_order",
            default="descending",
            help=(
                "Sort %s by start timestamp, "
                "asc|ascending (earliest first) "
                "or desc|descending (latest first)" % self.resource_name
            ),
        )

        # Filter options
        self.group.add_argument(
            "-c", "--trace-tag", help="Trace-tag to filter the list."
        )
        self.group.add_argument(
            "-e", "--execution", help="Execution to filter the list."
        )
        self.group.add_argument("-r", "--rule", help="Rule to filter the list.")
        self.group.add_argument(
            "-g", "--trigger-instance", help="TriggerInstance to filter the list."
        )
        # Display options
        self.parser.add_argument(
            "-a",
            "--attr",
            nargs="+",
            default=self.display_attributes,
            help=(
                "List of attributes to include in the "
                'output. "all" will return all '
                "attributes."
            ),
        )
        self.parser.add_argument(
            "-w",
            "--width",
            nargs="+",
            type=int,
            default=None,
            help=("Set the width of columns in output."),
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        # Filtering options
        if args.trace_tag:
            kwargs["trace_tag"] = args.trace_tag
        if args.trigger_instance:
            kwargs["trigger_instance"] = args.trigger_instance
        if args.execution:
            kwargs["execution"] = args.execution
        if args.rule:
            kwargs["rule"] = args.rule

        if args.sort_order:
            if args.sort_order in ["asc", "ascending"]:
                kwargs["sort_asc"] = True
            elif args.sort_order in ["desc", "descending"]:
                kwargs["sort_desc"] = True
        return self.manager.query_with_count(limit=args.last, **kwargs)

    def run_and_print(self, args, **kwargs):
        instances, count = self.run(args, **kwargs)

        if instances and len(instances) == 1:
            # For a single Trace we must include the components unless
            # user has overriden the attributes to display
            if args.attr == self.display_attributes:
                args.attr = ["all"]
            self.print_trace_details(trace=instances[0], args=args)

            if not args.json and not args.yaml:
                if args.last and count and count > args.last:
                    table.SingleRowTable.note_box(self.resource_name, 1)
        else:
            if args.json or args.yaml:
                self.print_output(
                    instances,
                    table.MultiColumnTable,
                    attributes=args.attr,
                    widths=args.width,
                    json=args.json,
                    yaml=args.yaml,
                    attribute_transform_functions=self.attribute_transform_functions,
                )
            else:
                self.print_output(
                    instances,
                    table.MultiColumnTable,
                    attributes=args.attr,
                    widths=args.width,
                    attribute_transform_functions=self.attribute_transform_functions,
                )

                if args.last and count and count > args.last:
                    table.SingleRowTable.note_box(self.resource_name, args.last)


class TraceGetCommand(resource.ResourceGetCommand, SingleTraceDisplayMixin):
    display_attributes = ["all"]
    attribute_display_order = TRACE_ATTRIBUTE_DISPLAY_ORDER
    attribute_transform_functions = {
        "start_timestamp": format_isodate_for_user_timezone
    }

    pk_argument_name = "id"

    def __init__(self, resource, *args, **kwargs):
        super(TraceGetCommand, self).__init__(resource, *args, **kwargs)

        # Causation chains
        self.causation_group = self.parser.add_mutually_exclusive_group()

        self.causation_group.add_argument(
            "-e", "--execution", help="Execution to show causation chain."
        )
        self.causation_group.add_argument(
            "-r", "--rule", help="Rule to show causation chain."
        )
        self.causation_group.add_argument(
            "-g", "--trigger-instance", help="TriggerInstance to show causation chain."
        )

        # display filter group
        self.display_filter_group = self.parser.add_argument_group()

        self.display_filter_group.add_argument(
            "--show-executions", action="store_true", help="Only show executions."
        )
        self.display_filter_group.add_argument(
            "--show-rules", action="store_true", help="Only show rules."
        )
        self.display_filter_group.add_argument(
            "--show-trigger-instances",
            action="store_true",
            help="Only show trigger instances.",
        )
        self.display_filter_group.add_argument(
            "-n",
            "--hide-noop-triggers",
            action="store_true",
            help="Hide noop trigger instances.",
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        resource_ids = getattr(args, self.pk_argument_name, None)
        resources = self._get_multiple_resources(
            resource_ids=resource_ids, kwargs=kwargs
        )
        return resources

    @resource.add_auth_token_to_kwargs_from_cli
    def run_and_print(self, args, **kwargs):
        traces = self.run(args, **kwargs)

        for trace in traces:
            # First filter for causation chains
            trace = self._filter_trace_components(trace=trace, args=args)
            # next filter for display purposes
            trace = self._apply_display_filters(trace=trace, args=args)
            self.print_trace_details(trace=trace, args=args)

    @staticmethod
    def _filter_trace_components(trace, args):
        """
        This function walks up the component causal chain. It only returns
        properties in the causal chain and nothing else.
        """
        # check if any filtering is desired
        if not (args.execution or args.rule or args.trigger_instance):
            return trace

        component_id = None
        component_type = None

        # pick the right component type
        if args.execution:
            component_id = args.execution
            component_type = "action_execution"
        elif args.rule:
            component_id = args.rule
            component_type = "rule"
        elif args.trigger_instance:
            component_id = args.trigger_instance
            component_type = "trigger_instance"

        # Initialize collection to use
        action_executions = []
        rules = []
        trigger_instances = []

        # setup flag to properly manage termination conditions
        search_target_found = component_id and component_type

        while search_target_found:
            components_list = []
            if component_type == "action_execution":
                components_list = trace.action_executions
                to_update_list = action_executions
            elif component_type == "rule":
                components_list = trace.rules
                to_update_list = rules
            elif component_type == "trigger_instance":
                components_list = trace.trigger_instances
                to_update_list = trigger_instances
            # Look for search_target in the right collection and
            # once found look up the caused_by to keep movig up
            # the chain.
            search_target_found = False
            # init to default value
            component_caused_by_id = None
            for component in components_list:
                test_id = component["object_id"]
                if test_id == component_id:
                    caused_by = component.get("caused_by", {})
                    component_id = caused_by.get("id", None)
                    component_type = caused_by.get("type", None)
                    # If provided the component_caused_by_id must match as well. This is mostly
                    # applicable for rules since the same rule may appear multiple times and can
                    # only be distinguished by causing TriggerInstance.
                    if (
                        component_caused_by_id
                        and component_caused_by_id != component_id
                    ):
                        continue
                    component_caused_by_id = None
                    to_update_list.append(component)
                    # In some cases the component_id and the causing component are combined to
                    # provide the complete causation chain. Think rule + triggerinstance
                    if component_id and ":" in component_id:
                        component_id_split = component_id.split(":")
                        component_id = component_id_split[0]
                        component_caused_by_id = component_id_split[1]
                    search_target_found = True
                    break

        trace.action_executions = action_executions
        trace.rules = rules
        trace.trigger_instances = trigger_instances
        return trace

    @staticmethod
    def _apply_display_filters(trace, args):
        """
        This function looks at the disaply filters to determine which components
        should be displayed.
        """
        # If all the filters are false nothing is to be filtered.
        all_component_types = not (
            args.show_executions or args.show_rules or args.show_trigger_instances
        )

        # check if noop_triggers are to be hidden. This check applies whenever TriggerInstances
        # are to be shown.
        if (
            all_component_types or args.show_trigger_instances
        ) and args.hide_noop_triggers:
            filtered_trigger_instances = []
            for trigger_instance in trace.trigger_instances:
                is_noop_trigger_instance = True
                for rule in trace.rules:
                    caused_by_id = rule.get("caused_by", {}).get("id", None)
                    if caused_by_id == trigger_instance["object_id"]:
                        is_noop_trigger_instance = False
                if not is_noop_trigger_instance:
                    filtered_trigger_instances.append(trigger_instance)
            trace.trigger_instances = filtered_trigger_instances

        if all_component_types:
            return trace

        if not args.show_executions:
            trace.action_executions = []

        if not args.show_rules:
            trace.rules = []

        if not args.show_trigger_instances:
            trace.trigger_instances = []

        return trace



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from typing import List

import copy

import six

__all__ = ["merge_dicts", "reencode_list_with_surrogate_escape_sequences"]


def merge_dicts(d1, d2):
    """
    Merge values from d2 into d1 ignoring empty / None values.

    :type d1: ``dict``
    :type d2: ``dict``

    :rtype: ``dict``
    """
    result = copy.deepcopy(d1)

    for key, value in six.iteritems(d2):
        if isinstance(value, dict):
            result[key] = merge_dicts(result[key], value)
        elif key not in result or value is not None:
            result[key] = value

    return result


def reencode_list_with_surrogate_escape_sequences(value: List[str]) -> List[str]:
    """
    Function which reencodes each item in the provided list replacing unicode surrogate escape
    sequences using actual unicode values.
    """
    result = []

    for item in value:
        try:
            item = item.encode("ascii", "surrogateescape").decode("utf-8")
        except UnicodeEncodeError:
            # Already a unicode string, nothing to do
            pass

        result.append(item)

    return result



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import argparse

from tests import base

from st2client.commands import trace as trace_commands
from st2client.models import trace as trace_models


class TraceCommandTestCase(base.BaseCLITestCase):
    def test_trace_get_filter_trace_components_executions(self):
        trace = trace_models.Trace()
        setattr(
            trace,
            "action_executions",
            [{"object_id": "e1", "caused_by": {"id": "r1:t1", "type": "rule"}}],
        )
        setattr(
            trace,
            "rules",
            [
                {
                    "object_id": "r1",
                    "caused_by": {"id": "t1", "type": "trigger_instance"},
                }
            ],
        )
        setattr(
            trace,
            "trigger_instances",
            [
                {"object_id": "t1", "caused_by": {}},
                {"object_id": "t2", "caused_by": {"id": "e1", "type": "execution"}},
            ],
        )

        args = argparse.Namespace()
        setattr(args, "execution", "e1")
        setattr(args, "show_executions", False)
        setattr(args, "show_rules", False)
        setattr(args, "show_trigger_instances", False)
        setattr(args, "hide_noop_triggers", False)

        trace = trace_commands.TraceGetCommand._filter_trace_components(trace, args)
        self.assertEqual(len(trace.action_executions), 1)
        self.assertEqual(len(trace.rules), 1)
        self.assertEqual(len(trace.trigger_instances), 1)

    def test_trace_get_filter_trace_components_rules(self):
        trace = trace_models.Trace()
        setattr(
            trace,
            "action_executions",
            [{"object_id": "e1", "caused_by": {"id": "r1:t1", "type": "rule"}}],
        )
        setattr(
            trace,
            "rules",
            [
                {
                    "object_id": "r1",
                    "caused_by": {"id": "t1", "type": "trigger_instance"},
                }
            ],
        )
        setattr(
            trace,
            "trigger_instances",
            [
                {"object_id": "t1", "caused_by": {}},
                {"object_id": "t2", "caused_by": {"id": "e1", "type": "execution"}},
            ],
        )

        args = argparse.Namespace()
        setattr(args, "execution", None)
        setattr(args, "rule", "r1")
        setattr(args, "trigger_instance", None)
        setattr(args, "show_executions", False)
        setattr(args, "show_rules", False)
        setattr(args, "show_trigger_instances", False)
        setattr(args, "hide_noop_triggers", False)

        trace = trace_commands.TraceGetCommand._filter_trace_components(trace, args)
        self.assertEqual(len(trace.action_executions), 0)
        self.assertEqual(len(trace.rules), 1)
        self.assertEqual(len(trace.trigger_instances), 1)

    def test_trace_get_filter_trace_components_trigger_instances(self):
        trace = trace_models.Trace()
        setattr(
            trace,
            "action_executions",
            [{"object_id": "e1", "caused_by": {"id": "r1:t1", "type": "rule"}}],
        )
        setattr(
            trace,
            "rules",
            [
                {
                    "object_id": "r1",
                    "caused_by": {"id": "t1", "type": "trigger_instance"},
                }
            ],
        )
        setattr(
            trace,
            "trigger_instances",
            [
                {"object_id": "t1", "caused_by": {}},
                {"object_id": "t2", "caused_by": {"id": "e1", "type": "execution"}},
            ],
        )

        args = argparse.Namespace()
        setattr(args, "execution", None)
        setattr(args, "rule", None)
        setattr(args, "trigger_instance", "t1")
        setattr(args, "show_executions", False)
        setattr(args, "show_rules", False)
        setattr(args, "show_trigger_instances", False)
        setattr(args, "hide_noop_triggers", False)

        trace = trace_commands.TraceGetCommand._filter_trace_components(trace, args)
        self.assertEqual(len(trace.action_executions), 0)
        self.assertEqual(len(trace.rules), 0)
        self.assertEqual(len(trace.trigger_instances), 1)

    def test_trace_get_apply_display_filters_show_executions(self):
        trace = trace_models.Trace()
        setattr(trace, "action_executions", ["1"])
        setattr(trace, "rules", ["1"])
        setattr(trace, "trigger_instances", ["1"])

        args = argparse.Namespace()
        setattr(args, "show_executions", True)
        setattr(args, "show_rules", False)
        setattr(args, "show_trigger_instances", False)
        setattr(args, "hide_noop_triggers", False)

        trace = trace_commands.TraceGetCommand._apply_display_filters(trace, args)
        self.assertTrue(trace.action_executions)
        self.assertFalse(trace.rules)
        self.assertFalse(trace.trigger_instances)

    def test_trace_get_apply_display_filters_show_rules(self):
        trace = trace_models.Trace()
        setattr(trace, "action_executions", ["1"])
        setattr(trace, "rules", ["1"])
        setattr(trace, "trigger_instances", ["1"])

        args = argparse.Namespace()
        setattr(args, "show_executions", False)
        setattr(args, "show_rules", True)
        setattr(args, "show_trigger_instances", False)
        setattr(args, "hide_noop_triggers", False)

        trace = trace_commands.TraceGetCommand._apply_display_filters(trace, args)
        self.assertFalse(trace.action_executions)
        self.assertTrue(trace.rules)
        self.assertFalse(trace.trigger_instances)

    def test_trace_get_apply_display_filters_show_trigger_instances(self):
        trace = trace_models.Trace()
        setattr(trace, "action_executions", ["1"])
        setattr(trace, "rules", ["1"])
        setattr(trace, "trigger_instances", ["1"])

        args = argparse.Namespace()
        setattr(args, "show_executions", False)
        setattr(args, "show_rules", False)
        setattr(args, "show_trigger_instances", True)
        setattr(args, "hide_noop_triggers", False)

        trace = trace_commands.TraceGetCommand._apply_display_filters(trace, args)
        self.assertFalse(trace.action_executions)
        self.assertFalse(trace.rules)
        self.assertTrue(trace.trigger_instances)

    def test_trace_get_apply_display_filters_show_multiple(self):
        trace = trace_models.Trace()
        setattr(trace, "action_executions", ["1"])
        setattr(trace, "rules", ["1"])
        setattr(trace, "trigger_instances", ["1"])

        args = argparse.Namespace()
        setattr(args, "show_executions", True)
        setattr(args, "show_rules", True)
        setattr(args, "show_trigger_instances", False)
        setattr(args, "hide_noop_triggers", False)

        trace = trace_commands.TraceGetCommand._apply_display_filters(trace, args)
        self.assertTrue(trace.action_executions)
        self.assertTrue(trace.rules)
        self.assertFalse(trace.trigger_instances)

    def test_trace_get_apply_display_filters_show_all(self):
        trace = trace_models.Trace()
        setattr(trace, "action_executions", ["1"])
        setattr(trace, "rules", ["1"])
        setattr(trace, "trigger_instances", ["1"])

        args = argparse.Namespace()
        setattr(args, "show_executions", False)
        setattr(args, "show_rules", False)
        setattr(args, "show_trigger_instances", False)
        setattr(args, "hide_noop_triggers", False)

        trace = trace_commands.TraceGetCommand._apply_display_filters(trace, args)
        self.assertEqual(len(trace.action_executions), 1)
        self.assertEqual(len(trace.rules), 1)
        self.assertEqual(len(trace.trigger_instances), 1)

    def test_trace_get_apply_display_filters_hide_noop(self):
        trace = trace_models.Trace()
        setattr(
            trace,
            "action_executions",
            [{"object_id": "e1", "caused_by": {"id": "r1:t1", "type": "rule"}}],
        )
        setattr(
            trace,
            "rules",
            [
                {
                    "object_id": "r1",
                    "caused_by": {"id": "t1", "type": "trigger_instance"},
                }
            ],
        )
        setattr(
            trace,
            "trigger_instances",
            [
                {"object_id": "t1", "caused_by": {}},
                {"object_id": "t2", "caused_by": {"id": "e1", "type": "execution"}},
            ],
        )

        args = argparse.Namespace()
        setattr(args, "show_executions", False)
        setattr(args, "show_rules", False)
        setattr(args, "show_trigger_instances", False)
        setattr(args, "hide_noop_triggers", True)

        trace = trace_commands.TraceGetCommand._apply_display_filters(trace, args)
        self.assertEqual(len(trace.action_executions), 1)
        self.assertEqual(len(trace.rules), 1)
        self.assertEqual(len(trace.trigger_instances), 1)



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import os
import json
import logging
import time
import calendar
import traceback
import platform

import six
import requests

# pylint: disable=import-error
from requests.packages.urllib3.exceptions import InsecureRequestWarning

from st2client import models
from st2client.config_parser import CLIConfigParser
from st2client.config_parser import ST2_CONFIG_DIRECTORY
from st2client.config_parser import ST2_CONFIG_PATH
from st2client.client import Client
from st2client.config import get_config
from st2client.utils.date import parse as parse_isotime
from st2client.utils.misc import merge_dicts

__all__ = ["BaseCLIApp"]

# Add Plattform Check to fix the Issue that PWD not exist on Windows and so the ST2 CLI not working.
# https://docs.python.org/3.8/library/pwd.html
if platform.system() != "Windows":
    import pwd

    # Fix for "os.getlogin()) OSError: [Errno 2] No such file or directory"
    os.getlogin = lambda: pwd.getpwuid(os.getuid())[0]

# How many seconds before the token actual expiration date we should consider the token as
# expired. This is used to prevent the operation from failing durig the API request because the
# token was just about to expire.
TOKEN_EXPIRATION_GRACE_PERIOD_SECONDS = 15

CONFIG_OPTION_TO_CLIENT_KWARGS_MAP = {
    "base_url": ["general", "base_url"],
    "auth_url": ["auth", "url"],
    "stream_url": ["stream", "url"],
    "api_url": ["api", "url"],
    "api_version": ["general", "api_version"],
    "api_key": ["credentials", "api_key"],
    "cacert": ["general", "cacert"],
    "debug": ["cli", "debug"],
    "basic_auth": ["credentials", "basic_auth"],
}


class BaseCLIApp(object):
    """
    Base class for StackStorm CLI apps.
    """

    LOG = logging.getLogger(__name__)  # logger instance to use
    client = None  # st2client instance

    # A list of command classes for which automatic authentication should be skipped.
    SKIP_AUTH_CLASSES = []

    def get_client(self, args, debug=False):
        ST2_CLI_SKIP_CONFIG = os.environ.get("ST2_CLI_SKIP_CONFIG", 0)
        ST2_CLI_SKIP_CONFIG = int(ST2_CLI_SKIP_CONFIG)

        skip_config = args.skip_config
        skip_config = skip_config or ST2_CLI_SKIP_CONFIG

        # Note: Options provided as the CLI argument have the highest precedence
        # Precedence order: cli arguments > environment variables > rc file variables
        cli_options = [
            "base_url",
            "auth_url",
            "api_url",
            "stream_url",
            "api_version",
            "cacert",
            "basic_auth",
        ]
        cli_options = {opt: getattr(args, opt, None) for opt in cli_options}
        if cli_options.get("cacert", None) is not None:
            if cli_options["cacert"].lower() in ["true", "1", "t", "y", "yes"]:
                cli_options["cacert"] = True
            elif cli_options["cacert"].lower() in ["false", "0", "f", "no"]:
                cli_options["cacert"] = False
        config_file_options = self._get_config_file_options(args=args)

        kwargs = {}

        if not skip_config:
            # Config parsing is not skipped
            kwargs = merge_dicts(kwargs, config_file_options)

        kwargs = merge_dicts(kwargs, cli_options)
        kwargs["debug"] = debug

        client = Client(**kwargs)

        if skip_config:
            # Config parsing is skipped
            self.LOG.info("Skipping parsing CLI config")
            return client

        # Ok to use config at this point
        rc_config = get_config()

        # Silence SSL warnings
        silence_ssl_warnings = rc_config.get("general", {}).get(
            "silence_ssl_warnings", False
        )
        if silence_ssl_warnings:
            # pylint: disable=no-member
            requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

        # We skip automatic authentication for some commands such as auth
        try:
            command_class_name = args.func.__self__.__class__.__name__
        except Exception:
            command_class_name = None

        if command_class_name in self.SKIP_AUTH_CLASSES:
            return client

        # We also skip automatic authentication if token is provided via the environment variable
        # or as a command line argument
        env_var_token = os.environ.get("ST2_AUTH_TOKEN", None)
        cli_argument_token = getattr(args, "token", None)
        env_var_api_key = os.environ.get("ST2_API_KEY", None)
        cli_argument_api_key = getattr(args, "api_key", None)
        if (
            env_var_token
            or cli_argument_token
            or env_var_api_key
            or cli_argument_api_key
        ):
            return client

        # If credentials are provided in the CLI config use them and try to authenticate
        credentials = rc_config.get("credentials", {})
        username = credentials.get("username", None)
        password = credentials.get("password", None)
        cache_token = rc_config.get("cli", {}).get("cache_token", False)

        if username:
            # Credentials are provided, try to authenticate agaist the API
            try:
                token = self._get_auth_token(
                    client=client,
                    username=username,
                    password=password,
                    cache_token=cache_token,
                )
            except requests.exceptions.ConnectionError as e:
                self.LOG.warn(
                    "Auth API server is not available, skipping authentication."
                )
                self.LOG.exception(e)
                return client
            except Exception as e:
                print("Failed to authenticate with credentials provided in the config.")
                raise e
            client.token = token
            # TODO: Hack, refactor when splitting out the client
            os.environ["ST2_AUTH_TOKEN"] = token

        return client

    def _get_config_file_options(self, args, validate_config_permissions=False):
        """
        Parse the config and return kwargs which can be passed to the Client
        constructor.

        :rtype: ``dict``
        """
        rc_options = self._parse_config_file(
            args=args, validate_config_permissions=validate_config_permissions
        )
        result = {}
        for kwarg_name, (section, option) in six.iteritems(
            CONFIG_OPTION_TO_CLIENT_KWARGS_MAP
        ):
            result[kwarg_name] = rc_options.get(section, {}).get(option, None)

        return result

    def _parse_config_file(self, args, validate_config_permissions=False):
        config_file_path = self._get_config_file_path(args=args)

        parser = CLIConfigParser(
            config_file_path=config_file_path,
            validate_config_exists=False,
            validate_config_permissions=validate_config_permissions,
            log=self.LOG,
        )
        result = parser.parse()
        return result

    def _get_config_file_path(self, args):
        """
        Retrieve path to the CLI configuration file.

        :rtype: ``str``
        """
        path = os.environ.get("ST2_CONFIG_FILE", ST2_CONFIG_PATH)

        if args.config_file:
            path = args.config_file

        path = os.path.abspath(os.path.expanduser(path))
        if path != ST2_CONFIG_PATH and not os.path.isfile(path):
            raise ValueError('Config "%s" not found' % (path))

        return path

    def _get_auth_token(self, client, username, password, cache_token):
        """
        Retrieve a valid auth token.

        If caching is enabled, we will first try to retrieve cached token from a
        file system. If cached token is expired or not available, we will try to
        authenticate using the provided credentials and retrieve a new auth
        token.

        :rtype: ``str``
        """
        if cache_token:
            token = self._get_cached_auth_token(
                client=client, username=username, password=password
            )
        else:
            token = None
        if not token:
            # Token is either expired or not available
            token_obj = self._authenticate_and_retrieve_auth_token(
                client=client, username=username, password=password
            )

            self._cache_auth_token(token_obj=token_obj)
            token = token_obj.token

        return token

    def _get_cached_auth_token(self, client, username, password):
        """
        Retrieve cached auth token from the file in the config directory.

        :rtype: ``str``
        """
        if not os.path.isdir(ST2_CONFIG_DIRECTORY):
            os.makedirs(ST2_CONFIG_DIRECTORY, mode=0o2770)
            # os.makedirs straight up ignores the setgid bit, so we have to set
            # it manually
            os.chmod(ST2_CONFIG_DIRECTORY, 0o2770)

        cached_token_path = self._get_cached_token_path_for_user(username=username)

        if not os.access(ST2_CONFIG_DIRECTORY, os.R_OK):
            # We don't have read access to the file with a cached token
            message = (
                'Unable to retrieve cached token from "%s" (user %s doesn\'t have read '
                "access to the parent directory). Subsequent requests won't use a "
                "cached token meaning they may be slower."
                % (cached_token_path, os.getlogin())
            )
            self.LOG.warn(message)
            return None

        if not os.path.isfile(cached_token_path):
            return None

        if not os.access(cached_token_path, os.R_OK):
            # We don't have read access to the file with a cached token
            message = (
                'Unable to retrieve cached token from "%s" (user %s doesn\'t have read '
                "access to this file). Subsequent requests won't use a cached token "
                "meaning they may be slower." % (cached_token_path, os.getlogin())
            )
            self.LOG.warn(message)
            return None

        # Safety check for too permissive permissions
        file_st_mode = oct(os.stat(cached_token_path).st_mode & 0o777)
        others_st_mode = int(file_st_mode[-1])

        if others_st_mode >= 2:
            # Every user has access to this file which is dangerous
            message = (
                'Permissions (%s) for cached token file "%s" are too permissive. Please '
                "restrict the permissions and make sure only your own user can read "
                "from or write to the file." % (file_st_mode, cached_token_path)
            )
            self.LOG.warn(message)

        with open(cached_token_path) as fp:
            data = fp.read()

        try:
            data = json.loads(data)

            token = data["token"]
            expire_timestamp = data["expire_timestamp"]
        except Exception as e:
            msg = (
                'File "%s" with cached token is corrupted or invalid (%s). Please delete '
                " this file" % (cached_token_path, six.text_type(e))
            )
            raise ValueError(msg)

        now = int(time.time())
        if (expire_timestamp - TOKEN_EXPIRATION_GRACE_PERIOD_SECONDS) < now:
            self.LOG.debug(
                'Cached token from file "%s" has expired' % (cached_token_path)
            )
            # Token has expired
            return None

        self.LOG.debug('Using cached token from file "%s"' % (cached_token_path))
        return token

    def _cache_auth_token(self, token_obj):
        """
        Cache auth token in the config directory.

        :param token_obj: Token object.
        :type token_obj: ``object``
        """
        if not os.path.isdir(ST2_CONFIG_DIRECTORY):
            os.makedirs(ST2_CONFIG_DIRECTORY, mode=0o2770)
            # os.makedirs straight up ignores the setgid bit, so we have to set
            # it manually
            os.chmod(ST2_CONFIG_DIRECTORY, 0o2770)

        username = token_obj.user
        cached_token_path = self._get_cached_token_path_for_user(username=username)

        if not os.access(ST2_CONFIG_DIRECTORY, os.W_OK):
            # We don't have write access to the file with a cached token
            message = (
                'Unable to write token to "%s" (user %s doesn\'t have write '
                "access to the parent directory). Subsequent requests won't use a "
                "cached token meaning they may be slower."
                % (cached_token_path, os.getlogin())
            )
            self.LOG.warn(message)
            return None

        if os.path.isfile(cached_token_path) and not os.access(
            cached_token_path, os.W_OK
        ):
            # We don't have write access to the file with a cached token
            message = (
                'Unable to write token to "%s" (user %s doesn\'t have write '
                "access to this file). Subsequent requests won't use a "
                "cached token meaning they may be slower."
                % (cached_token_path, os.getlogin())
            )
            self.LOG.warn(message)
            return None

        token = token_obj.token
        expire_timestamp = parse_isotime(token_obj.expiry)
        expire_timestamp = calendar.timegm(expire_timestamp.timetuple())

        data = {}
        data["token"] = token
        data["expire_timestamp"] = expire_timestamp
        data = json.dumps(data)

        # Note: We explictly use fdopen instead of open + chmod to avoid a security issue.
        # open + chmod are two operations which means that during a short time frame (between
        # open and chmod) when file can potentially be read by other users if the default
        # permissions used during create allow that.
        fd = os.open(cached_token_path, os.O_WRONLY | os.O_CREAT, 0o660)
        with os.fdopen(fd, "w") as fp:
            fp.write(data)
        os.chmod(cached_token_path, 0o660)

        self.LOG.debug('Token has been cached in "%s"' % (cached_token_path))
        return True

    def _authenticate_and_retrieve_auth_token(self, client, username, password):
        manager = models.ResourceManager(
            models.Token,
            client.endpoints["auth"],
            cacert=client.cacert,
            debug=client.debug,
        )
        instance = models.Token()
        instance = manager.create(instance, auth=(username, password))
        return instance

    def _get_cached_token_path_for_user(self, username):
        """
        Retrieve cached token path for the provided username.
        """
        file_name = "token-%s" % (username)
        result = os.path.abspath(os.path.join(ST2_CONFIG_DIRECTORY, file_name))
        return result

    def _print_config(self, args):
        config = self._parse_config_file(args=args, validate_config_permissions=False)

        for section, options in six.iteritems(config):
            print("[%s]" % (section))

            for name, value in six.iteritems(options):
                print("%s = %s" % (name, value))

    def _print_debug_info(self, args):
        # Print client settings
        self._print_client_settings(args=args)

        # Print exception traceback
        traceback.print_exc()

    def _print_client_settings(self, args):
        client = self.client

        if not client:
            return

        config_file_path = self._get_config_file_path(args=args)

        print("CLI settings:")
        print("----------------")
        print("Config file path: %s" % (config_file_path))
        print("Client settings:")
        print("----------------")
        print("ST2_BASE_URL: %s" % (client.endpoints["base"]))
        print("ST2_AUTH_URL: %s" % (client.endpoints["auth"]))
        print("ST2_API_URL: %s" % (client.endpoints["api"]))
        print("ST2_STREAM_URL: %s" % (client.endpoints["stream"]))
        print("ST2_AUTH_TOKEN: %s" % (os.environ.get("ST2_AUTH_TOKEN")))
        print("")
        print("Proxy settings:")
        print("---------------")
        print(
            "HTTP_PROXY: %s"
            % (os.environ.get("http_proxy", os.environ.get("HTTP_PROXY", "")))
        )
        print(
            "HTTPS_PROXY: %s"
            % (os.environ.get("https_proxy", os.environ.get("HTTPS_PROXY", "")))
        )
        print("")
